from . import CommandMessage, RegisterCommandClass

# Other command message implementations.
class SubDaemonCommandMessage(CommandMessage):
    def Execute(self):
        from stuphos.runtime.registry import getObject
        mgr = getObject('SubDaemon::Manager')
        if mgr is not None:
            daemon = mgr.getDaemon(self.SUBDAEMON_NAME)
            if daemon is not None:
                if not daemon.isRunning():
                    daemon.startProcess([])

                # Now, communicate the message contained in payload to the server.

class JythonCommandMessage(SubDaemonCommandMessage):
    SUBDAEMON_NAME = 'jyCommandApp'

RegisterCommandClass('application/x-jython', JythonCommandMessage)



# INI Format.
from stuphos.etc import fromJsonFile, fromJsonString, toJsonFile, toJsonString

from configparser import ConfigParser, NoOptionError
def getZoneConfigParser():
    return ConfigParser(defaults = dict(handler = '', specials = ''))

def parseZoneInfo(zone):
    info = zone.split(':', 3)
    if len(info) > 0:
        if info[0].isdigit():
            nr = int(info[0])
            if len(info) > 1:
                if len(info) > 2:
                    return (nr, info[1], info[2])
                return (nr, info[1], None)
            return (nr, None, None)

def getSoftConfigOption(cfg, section, name, *args, **kwd):
    try: return cfg.get(section, name, *args, **kwd)
    except NoOptionError:
        pass

def parseZoneConfig(config):
    modules = []
    for section in config.sections():
        if section.lower().startswith('module'):
            # Parse module name.  Skip all other sections.
            if len(section) > 6:
                if section[6] != ':':
                    continue
                name = section[7:]
            else:
                name = ''

            package = getSoftConfigOption(config, section, 'package')
            handler = getSoftConfigOption(config, section, 'handler')
            specials = getSoftConfigOption(config, section, 'specials', raw = True)

            zones = {}
            for opt in config.options(section):
                if opt.lower().startswith('zone'):
                    # Parse zone order number.
                    if len(opt) > 4:
                        if opt[4] != '.':
                            continue
                        znr = opt[5:]
                        if not znr.isdigit():
                            continue
                        znr = int(znr)
                    else:
                        znr = 0

                    info = config.get(section, opt)
                    info = parseZoneInfo(info)
                    if info is not None:
                        zones[znr] = info

            # Some modules just define specprocs:
            ##    if not package:
            ##        # Warning!
            ##        continue

            modules.append((name, package, handler, specials, zones))

    return modules

def parseZoneConfigFromFile(config_file):
    config = getZoneConfigParser()
    config.read([config_file])
    return parseZoneConfig(config)

def parseZoneConfigFromString(string):
    from io import StringIO
    config = getZoneConfigParser()
    config.readfp(StringIO(string)) # , '<init-buf>')
    return parseZoneConfig(config)

# XML Format.
from xml.sax.handler import ContentHandler
from xml.sax import parse as parseXML, parseString as parseXMLString

class ZoneModuleXML(ContentHandler):
    # Mostly validating.
    def __init__(self):
        self.modules = None
        self.current_m = None
        self.current_z = None
        self.setting = None

    def startElementNS(self, name, qname, attrs):
        if qname is not None:
            print('ns:', qname)

        # Todo: rewrite to look more state-based.
        lname = str(name).lower()
        if self.modules is None:
            if name.lower() != 'world':
                raise NameError(name)

            self.modules = []
        elif lname == 'module':
            if self.current_m is not None:
                raise NameError(name)
            else:
                # Todo: validate attrs
                self.current_m = dict((str(k).lower(), str(v)) for \
                                      (k, v) in list(attrs.items()))
        elif lname == 'zone':
            if self.current_m is None:
                raise NameError(name)
            elif self.current_z is not None:
                raise NameError(name)
            else:
                # Todo: validate attrs
                self.current_z = dict((str(k).lower(), str(v)) \
                                      for (k, v) in list(attrs.items()))
        else:
            if lname in ('package', 'handler'):
                if self.current_m is None:
                    raise NameError(name)
                if self.current_z is not None:
                    raise NameError(name)
            elif lname in ('number', 'guid'):
                if self.current_z is None:
                    raise NameError(name)
            elif lname != 'name':
                raise NameError(name)

    def endElementNS(self, name, qname):
        if qname is not None:
            print('ns:', qname)

        lname = str(name).lower()
        if lname == 'world':
            if self.modules is None:
                raise NameError(name)
            if self.current_m is not None:
                raise NameError(name)
        elif lname == 'module':
            if self.current_m is None:
                raise NameError(name)
            if self.current_z is not None:
                raise NameError

            # Plug it in.
            m = self.current_m
            self.current_m = None

            self.modules.append((m.get('name'),
                                 m.get('package'),
                                 m.get('handler'),
                                 m.get('zones')))

        elif lname == 'zone':
            if self.current_z is None:
                raise NameError(name)

            nr = self.current_z.get('number')
            if nr is None:
                raise ValueError('Zone without number!')
            elif not nr.isdigit():
                raise TypeError('Zone number is not a digit: %r' % nr)
            nr = int(nr)

            # Todo: validate settings.
            try: zones = self.current_m['zones']
            except KeyError:
                zones = self.current_m['zones'] = {}

            z = self.current_z
            self.current_z = None

            zones[nr] = (nr,
                         z.get('guid'),
                         z.get('name'))

        elif lname in ('package', 'handler'):
            if self.current_m is None:
                raise NameError(name)
            if self.current_z is not None:
                raise NameError(name)

            self.current_m[lname] = self.setting
            self.setting = None
        elif lname in ('guid', 'number'):
            if self.current_z is None:
                raise NameError(name)

            self.current_z[lname] = self.setting
            self.setting = None
        elif lname == 'name':
            if self.current_z:
                self.current_z[lname] = self.setting
            elif self.current_m:
                self.current_m[lname] = self.setting

            self.setting = None
        else:
            raise NameError(name)

    def characters(self, content):
        self.setting = str(content).strip()

    def startElement(self, name, attrs):
        return self.startElementNS(name, None, attrs)
    def endElement(self, name):
        return self.endElementNS(name, None)

    # Irrelevent (just interesting to see what comes through):
    def startPrefixMapping(self, prefix, uri):
        print('start-prefix-mapping:', prefix, url)
    def endPrefixMapping(self, prefix):
        print('end-prefix-mapping:', prefix)
    def processingInstruction(self, target, data):
        print('processing-instruction:', target, data)
    def skippedEntity(self, name):
        print('skipped-entity:', name)

def parseZoneXMLFromFile(xml_file):
    handler = ZoneModuleXML()
    parseXML(open(xml_file), handler)
    return handler.modules

def parseZoneXMLFromString(string):
    handler = ZoneModuleXML()
    parseXMLString(string, handler)
    return handler.modules

# JSON Format.
def parseZoneJSONFromFile(json_file):
    return _validateFormat(fromJsonFile(json_file))

def parseZoneJSONFromString(string):
    return _validateFormat(fromJsonString(json_file))

# Python Format -- UNSAFE!
def parseZonePyFromFile(py_file):
    return _validateFormat(eval(open(py_file).read()))
def parseZonePyFromString(string):
    return _validateFormat(eval(string))

# YAML
def parseZoneYAMLFromFile(yaml_file):
    from yaml import load
    return _validateFormat(load(open(yaml_file)))
def parseZoneYAMLFromString(string):
    from yaml import loads
    return _validateFormat(yaml.loads(string))

# Pickle
def parseZonePickleFromFile(pickle_file):
    from pickle import load
    return _validateFormat(load(open(pickle_file)))
def parseZonePickleFromString(string):
    from pickle import loads
    return _validateFormat(loads(string))

def _validateFormat(modules):
    # Todo: validate formats with associative keynames.
    for m in modules:
        if len(m) != 4:
            raise IndexError('Expected 4 items in module (not %d)' % len(m))
        if not m[1]:
            raise ValueError('No package specified')

        for z in m[3]:
            if len(z) != 3:
                raise IndexError('Expected 3 items in zone info (not %d)' % len(z))
            if type(z[0]) is not int:
                raise TypeError('Expected integer zone number (item 0), not %r' % type(z[0]).__name__)

    return modules

# Output Formats.
def dumpToBuffer(dumper, modules, *args, **kwd):
    from io import StringIO
    buf = StringIO()
    dumper(modules, buf, *args, **kwd)
    return buf.getvalue()

# Config
def dumpZoneInfoConfigToString(info):
    return ':'.join(v is not None and str(v) or '' for v in info)
def dumpZoneModuleConfigToStream(module, stream):
    if module[0]:
        print('[Module:%s]' % module[0], file=stream)
    else:
        print('[Module]', file=stream)

    if module[1]:
        print('package = %s' % module[1], file=stream)
    if module[2]:
        print('handler = %s' % module[2], file=stream)

    z = list(module[3].values())
    for x in range(len(z)):
        print('zone.%d = %s' % (x, dumpZoneInfoConfigToString(z[x])), file=stream)
def dumpZoneModuleConfigToString(module):
    return dumpToBuffer(dumpZoneModuleConfigToStream, module)
def dumpZoneConfigToFile(modules, stream):
    for m in modules:
        dumpZoneModuleConfigToStream(m, stream)
def dumpZoneConfigToString(modules):
    return dumpToBuffer(dumpZoneConfigToFile, modules)

# XML
def dumpZoneInfoXMLToString(info, indent = ''):
    fields = ['Number="%d"' % info[0]]
    if info[2]:
        fields.append('Name="%s"' % info[2])
    if info[1]:
        fields.append('GUID="%s"' % info[1])

    return '%s<Zone %s />' % (indent, ' '.join(fields))
def dumpZoneModuleXMLToFile(module, stream, indent = ''):
    stream.write('%s<Module>\n' % indent)
    if module[0]:
        stream.write('%s    <Name>%s</Name>\n' % (indent, module[0]))
    if module[1]:
        stream.write('%s    <Package>%s</Package>\n' % (indent, module[1]))
    if module[2]:
        stream.write('%s    <Handler>%s</Handler>\n' % (indent, module[2]))

    indent1 = indent + '    '
    for zone in module[3].values():
        stream.write(dumpZoneInfoXMLToString(zone, indent = indent1))
        stream.write('\n')

    stream.write('%s</Module>\n' % indent)
def dumpZoneModuleXMLToString(module, indent = ''):
    return dumpToBuffer(dumpZoneModuleXMLToFile, module, indent = indent)
def dumpZoneXMLToFile(modules, stream, indent = ''):
    # XXX Designed for use with loader...
    print('<World xmlns="org/stuph/mud/zones/config">', file=stream)
    indent1 = indent + '    '
    for zone in modules:
        dumpZoneModuleXMLToFile(zone, stream, indent = indent1)
    print('</World>', file=stream)
def dumpZoneXMLToString(modules):
    return dumpToBuffer(dumpZoneXMLToFile, modules)

# JSON
def dumpZoneJSONToFile(modules, stream):
    toJsonFile(modules, stream, indent = 1)
def dumpZoneJSONToString(modules):
    return dumpToBuffer(dumpZoneJSONToFile, modules)

# Py
def dumpZonePyToFile(modules, stream):
    for m in modules:
        print('name:', m[0], file=stream)
        print('package:', m[1], file=stream)
        print('handler:', m[2], file=stream)
        for z in m[3].values():
            print('  nr:', z[0], file=stream)
            print('  guid:', z[1], file=stream)
            print('  name:', z[2], file=stream)
def dumpZonePyToString(modules):
    return dumpToBuffer(dumpZonePyToFile, modules)

# YAML
def dumpZoneYAMLToFile(modules, stream):
    from yaml import dump
    dump(modules, stream)
def dumpZoneYAMLToString(modules):
    return dumpToBuffer(dumpZoneYAMLToFile, modules)

# Pickle
def dumpZonePickleToFile(modules, stream):
    from pickle import dump
    dump(modules, stream)
def dumpZonePickleToString(modules):
    from pickle import dumps
    return dumps(modules)

FORMATS = dict(cfg =  dict(parseFile   = parseZoneConfigFromFile,
                           parseString = parseZoneConfigFromString,
                           dumpFile    = dumpZoneConfigToFile,
                           dumpString  = dumpZoneConfigToString),

               xml =  dict(parseFile   = parseZoneXMLFromFile,
                           parseString = parseZoneXMLFromString,
                           dumpFile    = dumpZoneXMLToFile,
                           dumpString  = dumpZoneXMLToString),

               json = dict(parseFile   = parseZoneJSONFromFile,
                           parseString = parseZoneJSONFromString,
                           dumpFile    = dumpZoneJSONToFile,
                           dumpString  = dumpZoneJSONToString),

               py   = dict(parseFile   = parseZonePyFromFile,
                           parseString = parseZonePyFromString,
                           dumpFile    = dumpZonePyToFile,
                           dumpString  = dumpZonePyToString),

               yaml = dict(parseFile   = parseZoneYAMLFromFile,
                           parseString = parseZoneYAMLFromString,
                           dumpFile    = dumpZoneYAMLToFile,
                           dumpString  = dumpZoneYAMLToString),

               pkl  = dict(parseFile   = parseZonePickleFromFile,
                           parseString = parseZonePickleFromString,
                           dumpFile    = dumpZonePickleToFile,
                           dumpString  = dumpZonePickleToString))

def _getZoneFormat(filename, format = None):
    if format is None:
        assert filename
        from os.path import splitext
        format = splitext(filename)[1][1:].lower()

    info = FORMATS.get(format)
    if info is None:
        raise NameError('Unknown format: %r' % format)

    return info

def parseZoneFile(filename, format = None):
    parser = _getZoneFormat(filename, format)['parseFile']
    return parser(filename)
def dumpZoneFile(modules, stream, format):
    dumper = _getZoneFormat(None, format)['dumpFile']
    return dumper(modules, stream)

# Test Bed.
def testMain(argv = None):
    from optparse import OptionParser
    from sys import stdout

    parser = OptionParser()
    parser.add_option('-o', '--output-format', default = 'py')
    parser.add_option('-i', '--input-format')

    (options, args) = parser.parse_args(argv)

    for filename in args:
        modules = parseZoneFile(filename, options.input_format)
        dumpZoneFile(modules, stdout, options.output_format)

# todo: implement OLC submodules that manage zone-based modules.
# Associates zone-based modules built-in to a databased user component with olc-permissions/naming scheme.

if __name__ == '__main__':
    testMain()



#!/usr/local/bin/python
'Classes in this module read and write Zifnab\'s "Searches" from file.'
# Written by Fraun - Nov 9th 2005 (Modelled after the objlib reader)
#   Todo:  Decide on name of module: schlib? searchlib?
#          Invoke the --options from WSGI
#          Implement a lint detector for tildes and hash marks in front of strings (possibly stripping them)

from .dblib import RecordReader, FormatError, makePrintable
from os import getenv
from sys import stdout
from os.path import join, exists, basename
from pprint import pprint
from linecache import getline
from sys import exc_info


class SearchReader(RecordReader):
	# A little customization.
	MALFORMED_FIRST_RECORD_FMT='Malformed first search record header %c%r (expected #<Room Virtual NR>)'
	VirtualNRName = 'roomVNR'

	# This overloads the name of the vno in the record.
	# def newRecord(self, vnr):
	#	'Start of new current search record'
	#	# Merge searches for same rooms!  XXX This isn't supported yet!
	#	r = super(RecordReader, self).newRecord(vnr)
	#	s = self.getRecordSet()
	#	a = s.get(vnr, [])
	#	a.append(r)
	#	s[vnr] = a

	def parseRecord(self):
		'Load record area into logical Python object.  Returns the next Room VNR parsed.'

		# Pass each field read to handler function
		kind, reset = list(map(int, self.readLine().split()))
		self.recordField('type',     kind)
		self.recordField('reset',    reset)

		self.recordField('intval',   list(map(int,   self.readLine().split())))
		self.recordField('longval',  list(map(int,  self.readLine().split())))

		self.recordField('cmd',      self.tildeString('command'))
		self.recordField('noargs',   self.tildeString('command-sans-args'))
		self.recordField('keywords', self.tildeString('keywords'))

		self.recordField('strval',   list(map(lambda n, rs=self.tildeString:rs('string%d'%(n+1)), range(8))))
		self.recordField('objVNR',   int(self.readLine()))

		line=self.readLine()
		if line:
			c=line[0]
			if c=='$':        # detected end of records
				return    # return None-value to terminate caller

			if c=='#':        # start of next record
				return int(line[1:])

			raise FormatError("Trailing search record: '%s'" % makePrintable(c))

		# No $ at terminating records at end of file
		return self.EOFNoDollar()


##  The following code connects the reader with the filesystem given a CLI.
##  Also output routines that regenerate the .sch file format.

## Whitespace in this format is significant (notice beginning and end of string)
SEARCH_RECORD_FMT=\
"""#%(roomVNR)d
%(type)d %(reset)d
%(ival0)d %(ival1)d %(ival2)d %(ival3)d %(ival4)d %(ival5)d
%(lval0)ld %(lval1)ld
%(cmd)s~
%(noargs)s~
%(keywords)s~
%(sval0)s~
%(sval1)s~
%(sval2)s~
%(sval3)s~
%(sval4)s~
%(sval5)s~
%(sval6)s~
%(sval7)s~
%(objVNR)d"""

def formatSearches(out, sch, fmt=SEARCH_RECORD_FMT):
	'''
	This routine takes a file-like object for printing, and a dictionary containing
	the search info for rooms in a zone as prepared by the SearchReader procedures.

	An optional third argument overrides the format used in printing.

	It then outputs the data in the format readable by StuphMUD and the SearchReader,
	and additionally, prints the final '$' after any-or-all search records in said
	dictionary are formatted.

	'''
	order=list(sch.keys())
	order.sort() # by roomVNR

	for k in order:
		# Format this search to the file.  Does some magic on the dict.
		## Duplicate:
		A=sch.get(k)
		assert A

		S={}
		S.update(A)

		## Rearrange internals according to SEARCH_RECORD_FMT:
		S['ival0'], S['ival1'], S['ival2'], S['ival3'], \
				S['ival4'], S['ival5'] = A['intval']

		S['sval0'], S['sval1'], S['sval2'], S['sval3'], \
				S['sval4'], S['sval5'], S['sval6'], S['sval7'] = A['strval']

		S['lval0'], S['lval1'] = A['longval']

		## Output:
		print(fmt%S, file=out)

	print('$', file=out)

# DEFAULT_SEARCH_PATH='world/sch'
DEFAULT_SEARCH_PATH='tmp_sch'
DEFAULT_INSTALL_PATHKEY='STUPH_PATH'
DEFAULT_CLOBBER_OPT=False

def writeSearches(sch, path, overwrite=DEFAULT_CLOBBER_OPT):
	'Calculates the lowest zone in the given dictionary keys and formats the searches to filed path.'
	assert sch # Need key-values to decipher lowest zone NR

	lo=list(sch.keys())
	lo.sort() # descending: lowest last
	lo=int(lo[-1])/100

	fn=join(path, '%d.sch'%lo)

	if not overwrite:
		assert not exists(fn), 'Use overwrite=True when calling writeSearches!!'

	# Write the file
	return formatSearches(open(fn, 'w'), sch)

def loadNamedFile(fn):
	# Todo: Auto-detect file type (.sch file or another format)
	##  This is a potentially significant function, because it should be able to
	#   read multiple file formats (the .sch file, a file containing a python expression)

	sch=SearchReader(fn)
	sch.parseFile()

	return sch.consumeResult()

def doPrintSummary(searchFile, level=1):
	"""
	Print a description of the search file.

	The optional <level> argument can be an integer with a value that, when greater,
	will print more summary information than if a lesser value is specified:

		At level 0: no information is shown (useful for testing format)
		At level 1: displays the number of searchable rooms in this file

	"""
	sch=loadNamedFile(searchFile)

	# If level is 0, then we print nothing (but succeed this function)
	if level==1:
		print('%s: OK (%d searchable rooms)' % (basename(searchFile), len(sch)))


def doPrintAsDict(searchFile):
	pprint(loadNamedFile(searchFile))

def resolveInstallPath(path):
	'Finds and verifies the path.'
	path = path or \
		DEFAULT_SEARCH_PATH or \
			__import__('os').getenv(DEFAULT_INSTALL_PATHKEY)

	assert path, '--install-path option must be specified if STUPH_PATH environment is not set!'
	# assert __import__('os').path.exists(path)

def doInstall(searchFile, installPath, clobberSetting=None):
	installPath  = resolveInstallPath(installPath)
	searches     = loadNamedFile(searchFile)

	if clobberSetting in (True, False):
		writeSearches(searches, path, overwrite=clobberSetting)
	else:
		writeSearches(searches, path)

def processCmdln(cmdln):
	# Some settings.
	Path     = None  # No setting.
	Clobber  = None  # No setting.

	# Iterators.
	i=0
	n=len(cmdln)

	while i<n:
		a=cmdln[i]
		i+=1 # Next argument.

		# Do install operation inline.
		if a in ('-i', '--install'):
			assert i<n, str(a)+' <needs a search file>'
			doInstall(cmdln[i], Path, clobberSetting=Clobber)

			i+=1 # Consume the argument following.
			continue

		# Decipher install path.
		if a in ('-p', '--path', '--install-path'):
			assert i<n, str(a)+' <needs a path>'
			Path=cmdln[i]

			i+=1 # Consume the argument following.
			continue

		if a.startswith('--path='):
			Path=a[7:]  # this can be empty
			continue

		if a.startswith('--install-path='):
			Path=a[15:] # this can be empty
			continue

		# Order that the search file be parsed and printed in its internal form.
		if a in ('-d', '--dict', '--as-dict'):
			assert i<n, str(a)+' <needs a search file>'
			doPrintAsDict(cmdln[i])

			i+=1 # Consume the argument following.
			continue

		# Order that the search file be parsed and then printed in the formatted form.
		if a in ('-f', '--dump', '--dump-format'):
			assert i<n, str(a)+' <needs a search file>'
			formatSearches(stdout, loadNamedFile(cmdln[i]))

			i+=1 # Consume the argument following.
			continue

		# Order that the search file be parsed and only a level of summary be displayed.
		if a in ('-s', '--summary'):
			assert i<n, str(a)+' <needs a search file>'
			doPrintSummary(cmdln[i])

			i+=1 # Consume the argument following.
			continue

		if a.startswith('--summary='):
			# Use this to set the summary level
			assert i<n, str(a)+' <needs a search file>'
			level=int(a[:10:])

			doPrintSummary(cmdln[i], level=level)

			i+=1 # Consume the argument following.
			continue

		# Change clobber setting (if installation can overwrite files)
		if a in ('-X', '--overwrite', '--clobber'):
			Clobber=True
			continue

		if a in ('--no-overwrite', '--no-clobber'):
			Clobber=False
			continue

		if a.startswith('--clobber=') or a.startswith('--overwrite='):
			c=a[a.find('=')+1:].lower()

			# Decipher clobber setting:
			if c in ('true', 'yes', 'on'):
				Clobber=True
			else:
				assert c in ('false', 'no', 'off'), 'Unknown clobber argument: %r'%a
				Clobber=False

			continue

def main(cmdlnopts=None):
	if cmdlnopts is None:
		import sys
		cmdlnopts=getattr(sys, 'argv', [])[1:]

	try:
		processCmdln(cmdlnopts)

	except: # AssertionError:
		e=exc_info()
		tb=e[2]
		e=e[1]

		# Find last frame
		while tb and tb.tb_next:
			tb=tb.tb_next

		f=tb.tb_frame

		print(str(e), getline(f.f_code.co_filename, f.f_lineno).strip())

useDebugger=lambda:getenv('DEBUG', 'NO').upper() not in ('NO', 'OFF', 'FALSE')

if __name__=='__main__':
	if useDebugger():
		from pdb import runcall
		runcall(main)

	else:
		main()


##  Here's a script that could be useful.  It requires a path to searchlib.py and wldlib.py.
##  Put it in the bin directory, and call it something like 'searcheck'
#
#	#!/bin/bash
#
#	# Configure these things:
#	pylibdata_path="$HOME/pystuph/python/tools/libdata"
#	python_bin="/usr/bin/python" # or $HOME/python/python
#
#	# This is the default directory this script looks in when given a zone number.
#	stuphlib_worldsch="$HOME/stuph/lib/world/sch"
#
#	# These things don't need configuration.
#	searchlib_script="$pylibdata_path/searchlib.py"
#	stuphlib_schsuffix=".sch"
#
#	# Export PYTHON_PATH for it to propagate into the python executable process.
#	export PYTHON_PATH=$pylibdata_path
#
#	for zone in $*
#	do
#	    if [ -r $zone ]
#	    then
#	        search_file=$zone
#	    else
#	        search_file="$stuphlib_worldsch/$zone$stuphlib_schsuffix"
#	    fi
#
#	    # The PYTHON_PATH environment is exported above.
#	    $python_bin $searchlib_script --summary $search_file
#	done
#
##  This is how you might use the script:
#
#    searcheck 112 113 562 570 mysearch.sch my-other-search-file stuph/lib/world/sch/*.sch
#




# Command-Line Interface Routines
# (Used for subsystems.)
# Todo: merge in special command stuff.  Also, this should probably be migrated to mud.player.commands
#
import sys
import shlex

from contextlib import contextmanager

NoArgv = object()

@contextmanager
def settingCmdln(args):
    # Must temporarily shift the given argv into system module so that optparse functions right.
    previous_argv = getattr(sys, 'argv', NoArgv)

    if isinstance(args, str):
        args = args.split()
    elif args is None:
        args = []
    else:
        args = list(args)

    sys.argv = [''] + args
    try: yield # should yield args..? sys.argv?
    finally:
        if previous_argv is NoArgv:
            del sys.argv
        else:
            sys.argv = previous_argv

def parseOptionsOverSystem(parser, args):
    with settingCmdln(args):
        return parser.parse_args(list(args))


def printOptionsWithPrognameOverSystem(parser, progname, fileObj = None):
    previous_argv = getattr(sys, 'argv', NoArgv)

    progname = str(progname)
    assert progname
    sys.argv = [progname]

    try: return parser.print_help(file = fileObj)
    finally:
        if previous_argv is NoArgv:
            del sys.argv
        else:
            sys.argv = previous_argv

# Why isn't this in tools?
def maxWidth(sequence, width):
    for n in sequence:
        n = len(n)
        if n > width:
            width = n

    return width

def getKeyword(kwd, name, default = None):
    try: value = kwd[name]
    except KeyError: return default
    else: del kwd[name]

    return value

def Option(*args, **kwd):
    # Returns a pair of argument aggregates suitable for passing as the
    # positional and keyword arguments to a function (add_option).
    return (args, kwd)

def createCmdlnParser(*options):
    from optparse import OptionParser
    parser = OptionParser()
    for (args, kwd) in options:
        parser.add_option(*args, **kwd)

    return parser

class Cmdln:
    def __init__(self, progName, *options, **kwd):
        self.progName = progName
        self.options = options
        self.parser = createCmdlnParser(*options)
        self.shlex = bool(kwd.get('shlex', False))

    class HelpExit(Exception):
        pass

    class Parsed:
        def __init__(self, cmdln, command, argstr, options, args):
            self.cmdln = cmdln
            self.command = command
            self.options = options
            self.args = args
            self.argstr = argstr

            self.nextArg = iter(self).__next__

        def __repr__(self):
            return 'Parsed-Command: %s [%s] %r' % \
                   (self.command, self.argstr, self.options)

        def help(self):
            return self.cmdln.help()

        @property
        def argstr_stripped(self):
            return self.argstr.strip() if self.argstr else ''

        def __iter__(self):
            for arg in self.args:
                yield arg

            while True:
                yield ''

        def __len__(self):
            return len(self.args)

        def halfChop(self, string = None):
            if string is None:
                string = self.argstr

            a = string.strip()
            i = a.find(' ')
            if i < 0:
                return (a, '')

            return (a[:i], a[i:].lstrip())

        def oneArg(self, string = None):
            return self.halfChop(self.argstr if string is None else string)[0]

    def parseCommand(self, cmd, argstr):
        if argstr:
            if self.shlex:
                argv = shlex.split(argstr)
            else:
                argv = argstr.split()
        else:
            argv = []

        return self.Parsed(self, cmd, argstr, *self.parseArgv(argv))

    def parseArgv(self, argv):
        try: return parseOptionsOverSystem(self.parser, argv)
        except SystemExit:
            raise self.HelpExit

    __call__ = parseArgv

    def help(self):
        from io import StringIO
        buf = StringIO()

        printOptionsWithPrognameOverSystem(self.parser, self.progName, buf)
        return buf.getvalue()



#from django.conf.urls.defaults import *
from django.conf.urls import *


# class urlApplicationConfiguration(object):
#     'This organizes all of the application-specific components.'

#     def __new__(self, *args, **kwd):
#         # This behavior may present problems for some derivations.
#         i = object.__new__(self)
#         i.__init__(*args, **kwd)
#         return i()


#     # This object provides a way to merge a urlconf for a dynamic install.
#     def __init__(self):
#         self.patterns = patterns('')
#         self.installer = self._installer(self)

#     class _installer(object):
#         def __init__(self, config):
#             self.config = config

#         def __getattr__(self, name):
#             try: return object.__getattribute__(self, name)
#             except AttributeError:
#                 return self.mergeInstaller(self, name)

#         class mergeInstaller:
#             def __init__(self, installer, name):
#                 self.installer = installer
#                 self.name = name

#             def __call__(self, *args, **kwd):
#                 return self.installer.config.installComponent(self.name, *args, **kwd)


#     def __call__(self, *args, **kwd):
#         self.load(self.installer, *args, **kwd)

#         return self.patterns

#     def interpretMerge(self, result):
#         if result is not None:
#             if isinstance(result, self.configuration):
#                 result = result()

#             if result is not None:
#                 self.patterns += result

#         return self
#     __lshift__ = interpretMerge

#     def installComponent(self, name, *args, **kwd):
#         proc = getattr(self, name)
#         result = proc(self, *args, **kwd)

#         return self.interpretMerge(result)

#     @property
#     def scopes(self):
#         def scopeName(cls):
#             try: name = cls.name
#             except AttributeError:
#                 name = None

#             if not name:
#                 name = cls.__name__

#                 for ending in ['ConfigurationOptional', 'Configuration']:
#                     if name.endswith(ending):
#                         name = name[:-len(ending)]
#                         break

#             return name


#         def scopeClasses():
#             # todo: provide more scopes for grouping them by name.
#             for name in dir(self):
#                 try: cls = getattr(self, name)
#                 except AttributeError: pass
#                 else:
#                     if isinstance(cls, ClassType):
#                         if issubclass(cls, self.configuration):
#                             yield cls


#         for cls in scopeClasses():
#             yield (scopeName(cls), cls)


#     def loadConfiguration(self, config):
#         scopes = dict(self.scopes)

#         try: pattern = config.pop('url')
#         except AttributeError: pass
#         else:
#             view = config.pop('view')

#             s = config.pop('scope')
#             cls = scopes.get(s, self.configuration)

#             if isinstance(view, basestring):
#                 view = view.split('.')

#             o = cls

#             for n in view:
#                 o = getattr(o, n)

#             return url(pattern, o, **config)


#     class configuration:
#         def __init__(self, mainConfig):
#             self.mainConfig = mainConfig

#         scope = ''
#         name = ''

#         def __call__(self):
#             p = patterns(self.scope)
#             p += list(self.entries)

#             return p

#         @property
#         def entries(self):
#             for a in self:
#                 if isinstance(a, tuple):
#                     yield url(*a)
#                 else:
#                     yield a

#         @property
#         def modules(self):
#             return self.mainConfig.modules


# component = urlApplicationConfiguration.configuration


# debugOn()
try:
	from phsite.network import embedded
	from phsite.network.urls import urlImport as urlApplicationImport

except ImportError:
	from stuphos import logException
	logException(traceback = True)
	embedded = None

# import embedded

# urlpatterns = patterns('',
#     # Apache -- do this if !embedded.installed
#     # (r'^stuph',      include('stuphos.web.application.urls')),

#     # Development/Embedded Server -- Default deferrence.
#     # XXX first of all, why would this have ever worked, that is was previously named wrong... :skip:
#     # Second of all, we are not using the formal django include operation, we're just merging
#     # the urlconf objects using our application urls configuration, as seen later.
#     # (r'^',        include('phsite.network.urls')),
# )

urlpatterns = urlApplicationImport()


# Admin -- !embedded.installed
##    from django.contrib import admin
##    admin.autodiscover()
##
##    from os.path import dirname, join as joinpath
##    ADMIN_MEDIA_DIR = joinpath(dirname(admin.__file__), 'static')
##    ADMIN_MEDIA_DIR = joinpath(ADMIN_MEDIA_DIR, 'admin')
##
##    urlpatterns += patterns('',
##        (r'^admin/doc/',   include('django.contrib.admindocs.urls')),
##        (r'^media/admin/(?P<path>.*)$',
##                           'django.views.static.serve',
##                           {'document_root': ADMIN_MEDIA_DIR}),
##
##        # (r'^admin/(.*)',   admin.site.root)
##        # Django CMS requires this (old form) for template namespaces:
##        (r'^admin/', include(admin.site.urls)), 
##    )

# Databrowse.
##    from django.contrib import databrowse
##
##    databrowse_site_root = databrowse.site.root

# from django.contrib.auth.decorators import login_required
# databrowse_site_root = login_required(databrowse_site_root)

##    urlpatterns += patterns('', (r'^databrowse/(.*)', databrowse_site_root))
##
##    from web.stuph import initializeSite
##    initializeSite(admin, databrowse)

# CMS
##    if not embedded.installed or True: # todo: remove or True
##        from django.conf import settings
##        urlpatterns += patterns('',
##            url(r'^cms/media/(?P<path>.*)$', 'django.views.static.serve',
##                {'document_root': settings.CMS_MEDIA_ROOT, 'show_indexes': True}),
##            url(r'^cms/', include('cms.urls'))
##        )

# Castalian.
# urlpatterns += patterns('', url(r'^cas', include('castalian')))



from stuphlib.textlib import ReadTildeString

def ReadHelpFile(fp):
    # Hash Mark (#) - separated records with tilde-strings:
    #   suggests use of FileReader, but without vnum syntax...
    linenr = 0
    while True:
        line = fp.readline()
        if line == '':
            # Expected a terminating '$', but...
            break

        linenr += 1
        line = line.strip()

        if line == '$':
            # Normal terminator.
            break

        if line != '#':
            # Just an empty hash mark for count_hash_records
            raise SyntaxError('Expected hash mark (#) or terminator ($), got empty line (line #%d)' % linenr)

        # Read help entry.
        keywords = fp.readline()
        if keywords == '':
            raise EOFError('Expected keywords (line #%d)' % (linenr))

        linenr += 1
        keywords = keywords.strip()
        if not keywords:
            raise SyntaxError('Expected keywords, not empty line (line #%d)' % linenr)

        minlevel = fp.readline()
        if keywords == '':
            raise EOFError('Expected minlevel (line #%d)' % (linenr))

        linenr += 1
        minlevel = minlevel.strip()
        try: minlevel = int(minlevel)
        except ValueError:
            raise SyntaxError('Expected minlevel, not %r (line #%d)' % (minlevel, linenr))

        (text, nrlines) = ReadTildeString(fp)
        linenr += nrlines

        yield dict(keywords = keywords,
                   minlevel = minlevel,
                   text = text)

def LoadHelpFile(fp):
    records = dict()
    for help_entry in ReadHelpFile(fp):
        keywords = help_entry.pop('keywords')
        records[keywords] = help_entry

    return records

def ReadHelpScreen(fp):
    return fl.read()

# Indexing.
from stuphlib.dblib import LocalIndex, INDEX_FILE

class HelpIndex:
    def iterindex(self, index):
        for n in self.open(index):
            n = n.strip()
            if n == '$':
                break

            yield n.strip()

    def openHelpFile(self, helpfile):
        return self.open(helpfile)

class HelpLocalIndex(HelpIndex, LocalIndex):
    pass

def LoadHelpIndex(help_base, index_file = None):
    index = HelpLocalIndex(help_base)
    if index_file is None:
        index_file = INDEX_FILE

    help_table = dict()
    for helpfile in index.iterindex(index_file):
        # By help savefile.
        help_table[helpfile] = LoadHelpFile(index.openHelpFile(helpfile))

    return help_table

# Inspection Front End.
def main(argv = None):
    from sys import argv
    if len(argv) == 2:
        help_table = LoadHelpIndex(argv[1])

        from code import InteractiveConsole as IC
        import readline

        IC(locals = dict(help_table = help_table)).interact(banner = '')

if __name__ == '__main__':
    main()



from stuphlib.binary import Record, BuildRecordClass, ReadFile
from stuphlib.rent import RentedItemData

HOUSE_PRIVATE = 0
HOUSE_OPEN    = 1

def ModeName(mode):
    return {HOUSE_PRIVATE : 'Private',
            HOUSE_OPEN    : 'Open'}.get(mode) \
            or 'Mode #%d' % mode

class HouseControlRec(Record, metaclass=BuildRecordClass):
    class Types:
        MAX_GUESTS = 10

        RoomVnum   = Record.Int
        GuestArray = Record.Array(Record.Long, MAX_GUESTS)

    Fields = [('vnum',          Types.RoomVnum   ), # vnum of this house
              ('atrium',        Types.RoomVnum   ), # vnum of atrium
              ('exit_num',      Record.ShInt     ), # direction of house's exit
              ('built_on',      Record.TimeT     ), # date this house was built
              ('mode',          Record.Int       ), # mode of ownership
              ('owner',         Record.Long      ), # idnum of house's owner
              ('num_of_guests', Record.Int       ), # how many guests for house
              ('guests',        Types.GuestArray ), # idnums of house's guests
              ('last_payment',  Record.TimeT     ), # date of last house payment
              ('max_item_save', Record.Long      ), # max number of items saved
              ('spare1',        Record.Long      ),
              ('spare2',        Record.Long      ),
              ('spare3',        Record.Long      ),
              ('spare4',        Record.Long      ),
              ('spare5',        Record.Long      ),
              ('spare6',        Record.Long      ),
              ('spare7',        Record.Long      )]

    def __str__(self):
        return HouseToString(self)

def LoadHouseFile(fp):
    return list(ReadFile(fp, RentedItemData))

# Report.
def HouseToString(house):
    return '\n'.join(['House:',
                      '  Room          : #%d' % house.vnum,
                      '  Owner         : #%d' % house.owner,
                      '  Atrium        : #%d' % house.atrium,
                      '  Build Date    : %s' % datestring(house.built_on),
                      '  Mode          : %s' % ModeName(house.mode),

                      '  Guests        : (%d)' % house.num_of_guests,
                      getGuests(house.guests),

                      '  Last Payment  : [%d] %s' % (house.last_payment,
                                                     datestring(house.last_payment)),

                      '  Max Item Save : %d' % house.max_item_save,
                      ''])

from datetime import datetime
def datestring(seconds):
    return datetime.fromtimestamp(seconds).ctime()

def getGuests(guests):
    return '  Guests        : %r' % guests
    return '\n    '.join('Idnum       : #%d' % idnum \
                         for idnum in guests),

Housefix = 'house'
def main(argv = None):
    from os.path import dirname, join as joinpath
    if argv is None:
        from sys import argv

    for filename in argv[1:]:
        housepath = joinpath(dirname(dirname(filename)), Housefix)
        for house in ReadFile(open(filename), HouseControlRec):
            print(house)

            housefile = joinpath(housepath, '%d.%s' % (house.vnum, Housefix))
            for item in LoadHouseFile(open(housefile)):
                print(item)

if __name__ == '__main__':
    main()



from .constants import *
from .binary import Record, BuildRecordClass
from .binary import ReadFile, ReadOneFromFile, ReadBytes

NUM_OF_BOARDS = 32
BOARD_FILE = (lambda n: 'boards/board.%s' % n)


class BoardInfo:
    def __init__(self, vnum, read_lvl, write_lvl, remove_lvl,
                 name, auto_cleanup):

        self.vnum = vnum
        self.read_lvl = read_lvl
        self.write_lvl = write_lvl
        self.remove_lvl = remove_lvl
        self.name = name
        self.filename = BOARD_FILE(name)
        self.auto_cleanup = auto_cleanup

    def __str__(self):
        return 'Board #%d (%s)' % (self.vnum, self.filename)
    __repr__ = __str__


_builtin_boards = [BoardInfo(1216, 0, 0, LVL_OVERSEER,
                             ('remort'), False),
                   BoardInfo(1221, LVL_AVATAR, LVL_AVATAR, LVL_OVERSEER,
                             ('buildergroup'), False),
                   BoardInfo(1223, 0, 0, LVL_OVERSEER,
                             ('newbie'), False),
                   BoardInfo(1224, LVL_AVATAR, LVL_AVATAR, LVL_OVERSEER,
                             ('area'), False),
                   BoardInfo(1225, 0, 0, LVL_OVERSEER,
                             ('quest'), False),
                   BoardInfo(1226, 0, 0, LVL_OVERSEER,
                             ('clan'), False),
                   BoardInfo(1227, LVL_AVATAR, LVL_AVATAR, LVL_OVERSEER,
                             ('todo'), False),
                   BoardInfo(1228, 0, 0, LVL_OVERSEER,
                             ('idea'), False),
                   BoardInfo(1229, LVL_OVERSEER, LVL_OVERSEER, LVL_IMPL,
                             ('admin'), False),
                   BoardInfo(1230, 0, 0, LVL_OVERSEER,
                             ('skills'), True),
                   BoardInfo(1231, 0, 0, LVL_OVERSEER,
                             ('social'), False),
                   BoardInfo(1233, 0, 0, LVL_OVERSEER,
                             ('player_relations'), False),
                   BoardInfo(1234, 0, LVL_AVATAR, LVL_IMPL,
                             ('zifnab'), True),
                   BoardInfo(1245, 0, LVL_AVATAR, LVL_IMPL,
                             ('coder'), False),
                   BoardInfo(1260, LVL_AVATAR, LVL_AVATAR, LVL_OVERSEER,
                             ('clangroup'), False),
                   BoardInfo(1261, LVL_AVATAR, LVL_AVATAR, LVL_OVERSEER,
                             ('questgroup'), False),
                   BoardInfo(1263, LVL_AVATAR, LVL_AVATAR, LVL_OVERSEER,
                             ('continent'), False),
                   BoardInfo(3097, LVL_AVATAR, LVL_DEITY, LVL_IMPL,
                             ('freeze'), True),
                   BoardInfo(3098, LVL_AVATAR, LVL_AVATAR, LVL_ARCH_GOD,
                             ('immort'), True),
                   BoardInfo(3099, 0, 0, LVL_OVERSEER,
                             ('mort'), True),
                   BoardInfo(8299, LVL_AVATAR, LVL_AVATAR, LVL_OVERSEER,
                             ('roleplay'), False),
                   BoardInfo(40700, 0, 0, LVL_GOD,
                             ('forsaken'), False),
                   BoardInfo(40767, 0, 0, LVL_DEMIGOD,
                             ('avalon'), False),
                   BoardInfo(40792, 0, 0, LVL_OVERSEER,
                             ('clan_council'), False),
                   BoardInfo(40797, 0, 0, LVL_OVERSEER,
                             ('mort_council'), False),
                   BoardInfo(40801, 0, 0, LVL_IMMORT,
                             ('gwar'), False),
                   BoardInfo(40832, 0, 0, LVL_IMMORT,
                             ('imperial_knights'), False),
                   BoardInfo(40902, 0, 0, LVL_IMMORT,
                             ('evil_empire'), False),
                   BoardInfo(41004, 0, 0, LVL_ETERNAL_GOD,
                             ('widowmaker'), False),
                   BoardInfo(41106, 0, 0, LVL_DEMIGOD,
                             ('valheru'), False),
                   BoardInfo(41131, 0, 0, LVL_IMMORT,
                             ('magic_roundabout'), False),
                   BoardInfo(41171, 0, 0, LVL_IMMORT,
                             ('ahroun'), False)]

class NumberOfMessages(Record, metaclass=BuildRecordClass):
    Fields = [('number', Record.Int)]

class BoardMessageInfo(Record, metaclass=BuildRecordClass):
    Fields = [('slot_num'   , Record.Int     ),
              # XXX This should be hacked ito a four-byte type
              ('heading'    , Record.Pointer ), # Unused
              ('level'      , Record.Int     ),
              ('heading_len', Record.Int     ),
              ('message_len', Record.Int     ),
              # XXX This should be hacked into a four-byte type
              ('birth'      , Record.TimeT   )]

class BoardMessageInfo_Compat(Record, metaclass=BuildRecordClass):
    # All of these fields assume files created on a 32-bit machine.
    Fields = [('slot_num'   , Record.Int     ),
              ('heading'    , Record.Int     ), # Unused
              ('level'      , Record.Int     ),
              ('heading_len', Record.Int     ),
              ('message_len', Record.Int     ),
              ('birth'      , Record.Int     )]

class BoardMessage:
    def __init__(self, info, header, message):
        self.info = info
        self.header = header
        self.message = message

    def __str__(self):
        return '%s\n%s' % (self.header, self.message)

def ReadBoardFile(fp, compat = False):
    nr = ReadOneFromFile(fp, NumberOfMessages)
    recordClass = BoardMessageInfo_Compat if compat else BoardMessageInfo
    for x in range(nr.number):
        msginfo = ReadOneFromFile(fp, recordClass)
        yield BoardMessage(msginfo,
                           header = ReadBytes(fp, msginfo.heading_len),
                           message = ReadBytes(fp, msginfo.message_len))

def LoadBuiltinBoards(base, compat = False):
    from os.path import join as joinpath
    for board in _builtin_boards:
        fp = open(joinpath(base, board.filename))
        messages = list(ReadBoardFile(fp, compat = compat))
        yield (board, messages)

BOARD_INFO_TEMPLATE = '''vnum: %d
read-level: %d
write-level: %d
remove-level: %d
auto-cleanup: %s
'''

def getBoardInfoString(board):
    return BOARD_INFO_TEMPLATE % \
           (board.vnum, board.read_lvl, board.write_lvl,
            board.remove_lvl, board.auto_cleanup)

from datetime import datetime
def getBoardMessageBoxString(messages):
    # return one string with all messages in multipart format.
    from io import StringIO
    buf = StringIO()

    for msg in messages:
        body = msg.message + '\n\n'

        print('Content-Length  : %d' % len(body), file=buf)
        print('X-Heading       : %s' % msg.header, file=buf)
        print('X-Level         : %d' % msg.info.level, file=buf)
        print('Date            : %s' % datetime.fromtimestamp(msg.info.birth).ctime(), file=buf)
        print('X-tra           : slot_num = %d' % msg.info.slot_num, file=buf)
        print(file=buf)
        buf.write(body)

    return buf.getvalue()

def main(argv = None):
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('-z', '--zip-file')
    parser.add_option('--compat', action = 'store_true')
    (options, args) = parser.parse_args(argv)

    if len(args) > 0:
        base = args[0]

        if options.zip_file:
            from zipfile import ZipFile
            zf = ZipFile(open(options.zip_file, 'w'), 'w')

        # import pdb; pdb.set_trace()

        for (board, messages) in LoadBuiltinBoards(base, compat = options.compat):
            if options.zip_file:
                zf.writestr('%s/INFO' % board.name, getBoardInfoString(board))
                zf.writestr('%s/MESSAGES' % board.name, getBoardMessageBoxString(messages))
            else:
                print(board)
                for msg in messages:
                    print(msg)

if __name__ == '__main__':
    main()



from stuphos.runtime.architecture import extensionObject
from ph.interpreter.mental.native import _safe_native

from stuphos.language.document.structural import parseStructuredMessage as parseMessage, StructuredEncoding

from stuphos.kernel import vmNewSequence

from json import dumps as toJsonString

__all__ = ['parseWm', 'typeWm', 'itemsWm']


class BuildingWMNode(extensionObject):
	@property
	def name(self):
		return self._object.name

	@property
	def typeName(self):
		return self._object.typeName

	@property
	def subtype(self):
		return self._object.subtype

	@property
	def value(self):
		return BuildingWM.wrap(self._object.value)


	def data(self, value = True):
		yield self.name
		yield self.typeName
		yield self.subtype

		if value:
			yield self.value

	def __iter__(self):
		return self.data(False)


	@property
	def definition(self):
		name = self.name
		name = toJsonString(name) if ('"' in name or "'" in name) else name

		typeName = self.typeName
		subtype = self.subtype

		if typeName:
			defn = f'({typeName}${subtype or ""})'
		elif subtype:
			defn = f'({subtype})'
		else:
			defn = ''

		return name + defn

	def __repr__(self):
		return f'<building {self.definition}>'


class BuildingWM(extensionObject):
	_parseMessage = parseMessage
	_buildingClass = StructuredEncoding.ClassMap.Building
	_buildingNodeClass = BuildingWMNode


	@classmethod
	def _parse(self, source, *args, **kwd):
		return self(self._parseMessage \
			(source, *args, **kwd))


	@property
	def _value(self):
		o = self._object
		return o.value if isinstance(o, self._buildingClass) else o


	def __len__(self):
		return len(self._value)
	def __str__(self):
		return str(self._value)

	def __iter__(self):
		o = self._value

		if isinstance(o, dict):
			return iter(vmNewSequence(o.keys()))

		return iter(o)

	def __getitem__(self, item):
		return self.__class__(self._value[item])

	def __getattr__(self, name):
		try: return object.__getattribute__(self, name)
		except AttributeError:
			return self[name]

	def __call__(self):
		if isinstance(self._object, self._buildingClass):
			return self._buildingNodeClass(self._object)

		raise TypeError(self._object.__class__)
	building = property(__call__)


@_safe_native
def parseWm(frame, source, *args, **kwd):
	source = 'WMC []\n\n' + str.__str__(source)
	return BuildingWM._parse(source, *args, **kwd)


@_safe_native
def typeWm(frame, node, default = None):
	if isinstance(node._object, BuildingWM._buildingClass) or \
		isinstance(node._object, BuildingWMNode):
		return 'building'

	if not isinstance(node, BuildingWM):
		raise TypeError(node.__class__)

	if isinstance(node._object, dict):
		return 'mapping'
	if isinstance(node._object, (list, tuple)):
		return 'sequence'
	if isinstance(node._object, str):
		return 'string'
	if isinstance(node._object, None):
		return 'none'
	if isinstance(node._object, bool):
		return 'bool'
	if isinstance(node._object, (int, float)):
		return 'number'

	return default

@_safe_native
def itemsWm(frame, node):
	if not isinstance(node, BuildingWM):
		raise TypeError(node.__class__)

	if not isinstance(node._object, dict):
		raise ValueError(node._object)

	return frame.task.sequence(node._object.items())



# def lookupSkill(name):
# 	import game
# 	game.syslog('lookup-skill: %r' % name)

# 	return name

spellTable = dict()

def manualSpell(caster, spell):
	try: callback = spellTable[spell.number]
	except KeyError: pass
	else:
		try: callback(spell, caster)
		except:
			from ph import logException
			logException(traceback = True)

class MaxSkills(ValueError):
	VALUE = 500000 # MAX_SPELLS

def nextAvailableSkillNr(name):
	import world

	for a in world.magic.all_abilities():
		if a.number == 0:
			continue
		if a.name == name: # todo: make this stronger
			return a.number
		if a.name == '!UNUSED!':
			return a.number

	# d = set(a.number for a in world.magic.all_abilities())

	# i = 0
	# while i < MaxSkills.VALUE:
	# 	if i in d:
	# 		i += 1
	# 		continue

	# 	return i

	raise MaxSkills()

def registerSpell(name, max_mana, min_mana, mana_change, minpos,
				  targets, violent, wearoff, callback):

	from world.magic import spello_manual

	nr = nextAvailableSkillNr(name)
	spello_manual(nr, name, max_mana, min_mana, mana_change,
				  minpos, targets, violent, wearoff)

	spellTable[nr] = callback

spello = registerSpell

class Spell:
	name = ''
	max_mana = 0
	min_mana = 0
	mana_change = 0
	minpos = 'standing'
	targets = []
	violent = False
	wearoff = ''

	def __call__(self):
		registerSpell(self.name, self.max_mana, self.min_mana, self.mana_change,
					  self.minpos, self.targets, self.violent, self.wearoff,
					  self.manual)
		return self

	def manual(self, spell, caster):
		pass

def ASPELL(name, max_mana, min_mana, mana_change, minpos, targets, violent, wearoff,
	       register = True):

	def makeSpell(manual):
		spell = Spell()
		spell.name = name
		spell.max_mana = max_mana
		spell.min_mana = min_mana
		spell.mana_change = mana_change
		spell.minpos = minpos
		spell.targets = targets
		spell.violent = violent
		spell.wearoff = wearoff

		if manual is not None:
			spell.manual = manual

		return spell() if register else spell

	return makeSpell

#@apply
#@apply
class flare(Spell):
	name = 'flare'
	mana_change = 20
	violent = True

	def manual(self, spell, caster):
		# cause elemental damage.
		pass



# PyUtiLib Runtime Architecture
# Note: this isn't in use yet.
from pyutilib.component.core import *

PluginGlobals.clear()
PluginGlobals.interface_registry.clear()

def getTriggerName(name):
    return 'on' + name[0].upper() + name[1:].lower()

class Bridge:
    class Event:
        # XXX All shared across instances of Bridge for single subclass.
        def __init__(self, triggerName = None, behavior = None):
            self.triggerName = triggerName
            self.behavior = behavior

        def install(self, bridge, module, name):
            # Configure this instance.
            self.bridge = bridge
            self.name = name

            if self.triggerName is None:
                self.triggerName = getTriggerName(name)

            # And put into bridge module space.
            setattr(module, name, self)

        def __call__(self, *args, **kwd):
            return self.bridge.callEvent(self, *args, **kwd)
        def __repr__(self):
            return '<%s: %s>' % (self.__class__.__name__, self.name)

    def __init__(self, module):
        self.listeners = ExtensionPoint(self.Interface)

        for (name, event) in self.__class__.__dict__.items():
            if isinstance(event, self.Event):
                event.install(self, module, name)

    def callEvent(self, event, *args, **kwd):
        for o in self.listeners:
            trigger = getattr(o, event.triggerName, None)
            if callable(trigger):
                # Todo: implement calling conventions (like EventController) via event.behavior.
                #
                # This probably means putting most of this code back into Event,
                # but exposing what it needs (listeners), too.
                #
                # Note: this doesn't pass a event controller object as first argument.
                trigger(*args, **kwd)


# MUD World-Bridge Runtime Implementation.
# Todo: eliminate some redundancy in declaration of the methods & their bindings.
# Todo: get rid of trigger-based naming convention.
class IWorldEvents(Interface):
    # Trigger-based naming conventions.
    # todo: finish programming in all of these?
    def onHeartbeat(secs, usecs):
        'Called whenever possible.'
    def onMovement(ch, direction, cost):
        'Called when mobiles move.'

class WorldComponent(Plugin):
    # Q: Not sure if this is proper because it should be 'abstract'?
    # Hmm, this gets registered in the PluginEnvironment's registry,
    # as do all subclasses of this (under the class name, so they can
    # not be the same!!)  Research this.
    implements(IWorldEvents)

class WorldBridge(Bridge):
    Interface = IWorldEvents

    # Names as called via extension bridge.
    heartbeat = Bridge.Event()
    movement = Bridge.Event()

def bootStart():
    # Install the world events onto the module for dispatch.
    from stuphos.system.api import game
    WorldBridge(game.bridgeModule())


# Example.
class Facility:
    @classmethod
    def create(self):
        # hmm, PluginEnvironment.registry makes this design unrealistic?
        self.Instrument()

class X(Facility):
    # Whereby Facility will automatically instance this as plugin.
    class Instrument(WorldComponent):
        def onHeartbeat(self, secs, usecs):
            print('heartbeat:', secs, usecs)
            #raise NotImplementedError
        def onMovement(self, ch, direction, cost):
            print('movement:', ch, direction, cost)
            #raise NotImplementedError

import __main__ as main
WorldBridge(main)

X.create()



'''
Contains all of the Circle/StuphMUD structures.

'''
 #-
 # Copyright (c) 2008 Clint Banis (hereby known as "The Author")
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
 # are met:
 # 1. Redistributions of source code must retain the above copyright
 #    notice, this list of conditions and the following disclaimer.
 # 2. Redistributions in binary form must reproduce the above copyright
 #    notice, this list of conditions and the following disclaimer in the
 #    documentation and/or other materials provided with the distribution.
 # 3. All advertising materials mentioning features or use of this software
 #    must display the following acknowledgement:
 #        This product includes software developed by The Author, Clint Banis.
 # 4. The name of The Author may not be used to endorse or promote products
 #    derived from this software without specific prior written permission.
 #
 # THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS
 # ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 # TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 # PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS
 # BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 # CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 # SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 # INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 # CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 #

from ctypes import c_int, c_char_p, c_char, c_byte, c_ubyte, c_float
from ctypes import c_void_p, c_short, c_ushort, c_long, c_ulong
from ctypes import Structure, cast, _Pointer as PointerType
from ctypes import POINTER, CFUNCTYPE, _CFuncPtr, _FUNCFLAG_CDECL

# C Typedefs
c_time_t = c_int
c_bool = c_byte
c_size_t = c_int
c_socket_t = c_int

c_int_p = POINTER(c_int)
c_int_pp = POINTER(c_int_p)

c_char_pp = POINTER(c_char_p)

FILE_P = c_void_p

#
# This is intended for use as a void return type for cdecl convention,
# assuming that void equates to a c_int on the stack.  It seems perfectly
# reasonable to use `None` in this capacity, but that only crashes the
# interpreter.
#
# XXX `None` *should* be a valid restype!  (according to examples)
# c_void = None
#
c_void = c_int

# Standard library types
FILE = c_void_p

class TimeValue(Structure):
    _fields_ = []

TimeValue.P = POINTER(TimeValue)
c_timeval = TimeValue

# Return Value Conversion Facility
    #
    #  Is this true?
    #   [There seems to be a matter of POINTER types, even to simple
    #    ctypes data types, being invalid result types (???)]
    #

_c_default_retval = c_int
_conversion_functype_cache = {}
def AUTOFUNCTYPE(data_type, *argtypes):
    'Automatically install conversion function for pointer data type.'
    try:
        return _conversion_functype_cache[(data_type, argtypes)]
    except:
        class ConversionFunctionType(_CFuncPtr):
            _data_type_ = data_type

            _argtypes_ = argtypes
            _restype_ = _c_default_retval
            _flags_ = _FUNCFLAG_CDECL

            def __init__(self, *args, **kwd):
                _CFuncPtr.__init__(self, *args, **kwd)

                # Override the restype on the instance.
                self.restype = self.__cast_data__

                # I can't decide whether or not to implement this:
                # self.data_type = self._data_type_

            def __cast_data__(self, address):
                # The conversion routine expects an integer address.
                return cast(address, self._data_type_)

        _conversion_functype_cache[(data_type, argtypes)] = ConversionFunctionType
        return ConversionFunctionType

# MUD Typedefs
RealNumber = c_int
VirtualNumber = c_int

generic_rnum = RealNumber
generic_vnum = VirtualNumber

room_rnum = RealNumber
room_vnum = VirtualNumber

obj_rnum = RealNumber
obj_vnum = VirtualNumber

mob_rnum = RealNumber
mob_vnum = VirtualNumber

zone_rnum = RealNumber
zone_vnum = VirtualNumber

# Should we support these?
# sh_int = c_short

# Constants
MAX_INPUT_LENGTH = 256
MAX_RAW_INPUT_LENGTH = 512
SMALL_BUFSIZE = 4096
MAX_NAME_LENGTH = 20
MAX_EMAIL_LENGTH = 80

# Todo: re-cast all types that are unsigned.  I forgot that `ctypes`
# offered unsigned varities of the primitives.

# MUD Entities
class DescriptorData(Structure):
    'descriptor_data'
    HOST_LENGTH = 40

    class TxtBlock(Structure):
        'txt_block'

    TxtBlock.P = POINTER(TxtBlock)
    TxtBlock._fields_ = \
        [('text', c_char_p),
         ('aliased', c_int),
         ('next', TxtBlock.P)]

    class TxtQueue(Structure):
        'txt_q'

    TxtQueue.P = POINTER(TxtQueue)

DescriptorData.P = POINTER(DescriptorData)

DescriptorData.TxtQueue._fields_ = \
    [('head', DescriptorData.TxtBlock.P),
     ('tail', DescriptorData.TxtBlock.P)]

class TimeData(Structure):
    'time_data'

    _fields_ = \
        [('birth', c_time_t),
         ('logon', c_time_t),
         ('played', c_int)]

class CharacterData(Structure):
    'char_data'

    NUM_WEARS = 10
    MAX_COLOR_SET = 10
    MAX_PWD_LENGTH = 10

    class PlayerData(Structure):
        'char_player_data'

    class AbilityData(Structure):
        'char_ability_data'

        _fields_ = [('str', c_byte),
                    ('str_add', c_byte),
                    ('intel', c_byte),
                    ('wis', c_byte),
                    ('dex', c_byte),
                    ('con', c_byte),
                    ('cha', c_byte)]

    class PointData(Structure):
        'char_point_data'
        POINT_TYPE = c_short

        _fields_ = [('mana', POINT_TYPE),
                    ('max_mana', POINT_TYPE),
                    ('hit', POINT_TYPE),
                    ('max_hit', POINT_TYPE),
                    ('move', POINT_TYPE),
                    ('max_move', POINT_TYPE),
                    ('deaths', POINT_TYPE),
                    ('mkills', POINT_TYPE),
                    ('pkills', POINT_TYPE),
                    ('dts', POINT_TYPE),
                    ('qpoints', POINT_TYPE),
                    ('armor', POINT_TYPE),
                    ('gold', c_int),
                    ('bank_gold', c_int),
                    ('exp', c_int),
                    ('hitroll', c_byte),
                    ('damroll', c_byte)]

    class CharSpecialData(Structure):
        'char_special_data'

        class Saved(Structure):
            'char_special_data_saved'

    CharSpecialData.P = POINTER(CharSpecialData)

    class PlayerSpecialData(Structure):
        'player_special_data_saved'

        class Saved(Structure):
            'player_special_data'

            MAX_SKILLS = 200
            MAX_TONGUE = 3

        class AliasData(Structure):
            'alias_data'

        AliasData.P = POINTER(AliasData)

    PlayerSpecialData.P = POINTER(PlayerSpecialData)

    class NpcSpecialData(Structure):
        'mob_special_data'
        DICE_TYPE = c_byte

        class NpcMemoryRecord(Structure):
            'memory_rec_struct'

        class NpcActionList(Structure):
            'action_list'

    class FollowType(Structure):
        'follow_type'

    class GroupType(Structure):
        'group_type'

    class AffectedType(Structure):
        'affected_type'

    AffectedType.P = POINTER(AffectedType)

    class FileU(Structure):
        'char_file_u'

        MAX_PWD_LENGTH = 10
        MAX_AFFECT = 32
        MAX_COLOR_SET = 32

    FileU.P = POINTER(FileU)

CharacterData.P = POINTER(CharacterData)
CharacterData.PP = POINTER(CharacterData.P)

SPECIAL_TYPE = CFUNCTYPE(c_int, CharacterData.P, c_void_p, c_int, c_char_p)

class ExtraDescrData(Structure):
    'extra_descr_data'

ExtraDescrData.P = POINTER(ExtraDescrData)

ExtraDescrData._fields_ = \
    [('keyword', c_char_p),
     ('description', c_char_p),
     ('next', ExtraDescrData.P)]

class ItemData(Structure):
    'obj_data'

    class FlagData(Structure):
        'obj_flag_data'

        _fields_ = \
            [('value', c_int),
             ('type_flag', c_byte),
             ('wear_flags', c_int),
             ('extra_flags', c_int),
             ('anti_flags', c_int),
             ('weight', c_int),
             ('cost', c_int),
             ('cost_per_day', c_int),
             ('timer', c_int),
             ('trap', c_int),
             ('bitvector', c_long)]

    class AffectedType(Structure):
        'obj_affected_type'

        _fields_ = \
            [('location', c_byte),
             ('modifier', c_byte)]

    AffectedType.P = POINTER(AffectedType)

ItemData.P = POINTER(ItemData)
ItemData.PP = POINTER(ItemData.P)

class ItemRentData(Structure):
    'obj_file_elem'
    MAX_OBJ_AFFECT = 6

    _fields_ = \
        [('item_number', obj_vnum),
         ('name', c_char_p * 128),
         ('description', c_char_p * 256),
         ('short_description', c_char_p * 128),
         ('locate', c_short),
         ('value', c_int * 4),
         ('item_type', c_int),
         ('extra_flags', c_int),
         ('anti_flags', c_int),
         ('wear_flags', c_int),
         ('weight', c_int),
         ('timer', c_int),
         ('bitvector', c_long),
         ('affected', ItemData.AffectedType * MAX_OBJ_AFFECT)]

ItemRentData.P = POINTER(ItemRentData)

class RentInfo(Structure):
    'rent_info'

    _fields_ = \
        [('time', c_int),
         ('rentcode', c_int),
         ('net_cost_per_diem', c_int),
         ('gold', c_int),
         ('account', c_int),
         ('nitems', c_int),
         ('spare0', c_int),
         ('spare1', c_int),
         ('spare2', c_int),
         ('spare3', c_int),
         ('spare4', c_int),
         ('spare5', c_int),
         ('spare6', c_int),
         ('spare7', c_int)]

ItemData._fields_ = \
    [('item_number', obj_vnum),
     ('in_room', room_rnum),
     ('obj_flags', ItemData.FlagData),
     ('affected', ItemData.AffectedType),
     ('name', c_char_p),
     ('description', c_char_p),
     ('short_description', c_char_p),
     ('action_description', c_char_p),
     ('ex_description', ExtraDescrData.P),
     ('affected_by', CharacterData.AffectedType.P),
     ('carried_by', CharacterData.P),
     ('worn_by', CharacterData.P),
     ('worn_on', c_short),
     ('was_worn_on', c_short),
     ('in_obj', ItemData.P),
     ('contains', ItemData.P),
     ('next_content', ItemData.P),
     ('next', ItemData.P),
     ('tried_wear', c_bool),
     ('py_handle', c_void_p)]

CharacterData.CharSpecialData.Saved._fields_ = \
    [('alignment', c_int),
     ('idnum', c_long),
     ('act', c_long),
     ('affected_by', c_long),
     ('apply_saving_throw', c_short * 5)]

CharacterData.CharSpecialData._fields_ = \
    [('afk_message', c_char_p),
     ('fighting', CharacterData.P),
     ('hunting', CharacterData.P),
     ('wielded_by', CharacterData.P),
     ('position', c_byte),
     ('carry_weight', c_int),
     ('carry_items', c_byte),
     ('timer', c_int),
     ('arena', c_int),
     ('tracking', c_int),
     ('track_timer', c_int),
     ('saved', CharacterData.CharSpecialData.Saved)]

CharacterData.PlayerData._fields_ = \
    [('passwd', c_byte * (CharacterData.MAX_PWD_LENGTH+1)),
     ('name', c_char_p),
     ('short_descr', c_char_p),
     ('long_descr', c_char_p),
     ('description', c_char_p),
     ('title', c_char_p),
     ('nickname', c_char_p),
     ('prename', c_char_p),
     ('wizname', c_char_p),
     ('plan', c_char_p),
     ('email', c_char_p),
     ('remort', c_int),
     ('race', c_int),
     ('clan', c_int),
     ('clanrank', c_int),
     ('save_room', c_int),
     ('page_length', c_int),
     ('breath', c_int),
     ('sex', c_byte),
     ('chclass', c_byte),
     ('level', c_byte),
     ('hometown', c_int),
     ('time', TimeData),
     ('weight', c_ubyte),
     ('height', c_ubyte),
     ('login_time', c_time_t)]

class LastBuffer(Structure):
    'last_buffer'
    CHANNEL_BUFSIZE = 10

    _fields_ = \
        [('message', c_char_p),
         ('name', c_char_p),
         ('invis', c_int),
         ('when', c_time_t)]

LastBuffer.P  = POINTER(LastBuffer)
LastBuffer.PP = POINTER(LastBuffer.P)

CharacterData.PlayerSpecialData.Saved._fields_ = \
    [('skills', c_byte * (CharacterData.PlayerSpecialData.Saved.MAX_SKILLS+1)),
     ('PADDING0', c_byte),
     ('talks', c_bool * CharacterData.PlayerSpecialData.Saved.MAX_TONGUE),
     ('wimp_level', c_int),
     ('freeze_level', c_byte),
     ('invis_level', c_short),
     ('load_room', room_vnum),
     ('pref', c_long),
     ('pref2', c_long),
     ('disp', c_long),
     ('bad_pws', c_byte),
     ('conditions', c_byte * 3),
     ('spells_to_learn', c_int),
     ('marital_status', c_int),
     ('num_marriages', c_int),
     ('married_to', c_long),
     ('spell_attack', c_int * 2),
     ('remort_points', c_int)]

CharacterData.PlayerSpecialData.AliasData._fields_ = \
    [('alias', c_char_p),
     ('replacement', c_char_p),
     ('type', c_int),
     ('next', CharacterData.PlayerSpecialData.AliasData.P)]

CharacterData.PlayerSpecialData._fields_ = \
    [('saved', CharacterData.PlayerSpecialData.Saved),
     ('poofin', c_char_p),
     ('poofout', c_char_p),
     ('aliases', CharacterData.PlayerSpecialData.AliasData.P),
     ('last_tell', c_long),
     ('last_olc_targ', c_void_p),
     ('last_olc_mode', c_int),
     ('tell_buf', LastBuffer * LastBuffer.CHANNEL_BUFSIZE),
     ('ctell_buf', LastBuffer * LastBuffer.CHANNEL_BUFSIZE),
     ('wiznet_buf', LastBuffer * LastBuffer.CHANNEL_BUFSIZE),
     ('cmdlock', c_int)]

CharacterData.NpcSpecialData.NpcMemoryRecord._fields_ = \
    [('id', c_long),
     ('next', POINTER(CharacterData.NpcSpecialData.NpcMemoryRecord))]

CharacterData.NpcSpecialData.NpcActionList._fields_ = \
    [('type', c_byte),
     ('chance', c_int),
     ('spell', c_int),
     ('self', c_byte),
     ('cmd', c_char_p),
     ('next', POINTER(CharacterData.NpcSpecialData.NpcActionList))]

CharacterData.NpcSpecialData._fields_ = \
    [('last_direction', c_byte),
     ('memory', POINTER(CharacterData.NpcSpecialData.NpcMemoryRecord)),
     ('attack_type', c_int),
     ('walk_type', c_int),
     ('default_pos', c_byte),
     ('damnodice', CharacterData.NpcSpecialData.DICE_TYPE),
     ('damsizedice', CharacterData.NpcSpecialData.DICE_TYPE),
     ('gen_spec_act', POINTER(CharacterData.NpcSpecialData.NpcActionList)),
     ('timer', c_int)]

CharacterData.FollowType._fields_ = \
    [('follower', CharacterData.P),
     ('next', POINTER(CharacterData.FollowType))]

CharacterData.GroupType._fields_ = \
    [('member', CharacterData.P * 9),
     ('formation', c_int * 9)]

CharacterData.AffectedType._field_ = \
    [('type', c_short),
     ('duration', c_short),
     ('modifier', c_byte),
     ('location', c_byte),
     ('bitvector', c_long),
     ('next', CharacterData.AffectedType.P)]

CharacterData._fields_ = \
    [('pfilepos', c_int),
     ('nr', mob_rnum),
     ('in_room', room_rnum),
     ('was_in_room', room_rnum),
     ('wait', c_int),
     ('player', CharacterData.PlayerData),
     ('real_abils', CharacterData.AbilityData),
     ('aff_abils', CharacterData.AbilityData),
     ('points', CharacterData.PointData),
     ('char_specials', CharacterData.CharSpecialData),
     ('player_specials', CharacterData.PlayerSpecialData.P),
     ('mob_specials', CharacterData.NpcSpecialData),
     ('affected', POINTER(CharacterData.AffectedType)),
     ('equipment', POINTER(ItemData) * CharacterData.NUM_WEARS),
     ('color', c_int * CharacterData.MAX_COLOR_SET),
     ('carrying', POINTER(ItemData)),
     ('desc', POINTER(DescriptorData)),
     ('next_in_room', POINTER(CharacterData)),
     ('next', POINTER(CharacterData)),
     ('next_fighting', POINTER(CharacterData)),
     ('followers', POINTER(CharacterData.FollowType)),
     ('master', POINTER(CharacterData)),
     ('group', CharacterData.GroupType),
     ('holding', POINTER(CharacterData)),
     ('held_by', POINTER(CharacterData)),
     ('mount', POINTER(CharacterData)),
     ('mounted_by', POINTER(CharacterData)),
     ('py_handle', c_void_p)]

CharacterData.FileU._fields_ = \
    [('name', c_char * (MAX_NAME_LENGTH + 1)),
     ('email', c_char * (MAX_EMAIL_LENGTH + 1)),
     ('remort', c_int),
     ('race', c_int),
     ('clan', c_int),
     ('clanrank', c_int),
     ('page_length', c_int),
     ('save_room', c_int),
     ('sex', c_byte),
     ('chclass', c_byte),
     ('level', c_byte),
     ('hometown', c_short),
     ('birth', c_time_t),
     ('played', c_int),
     ('weight', c_ubyte),
     ('height', c_ubyte),
     ('pwd', c_char * (CharacterData.FileU.MAX_PWD_LENGTH + 1)),
     ('char_specials_saved', CharacterData.CharSpecialData.Saved),
     ('player_specials_saved', CharacterData.PlayerSpecialData.Saved),
     ('abilities', CharacterData.AbilityData),
     ('points', CharacterData.PointData),
     ('affected', CharacterData.AffectedType * CharacterData.FileU.MAX_AFFECT),
     ('color', c_int * CharacterData.FileU.MAX_COLOR_SET),
     ('last_login', c_time_t),
     ('last_logon', c_time_t),
     ('host', c_char * (DescriptorData.HOST_LENGTH + 1))]

class RoomData(Structure):
    'room_data'

    NUM_OF_DIRS = 6

    class DirectionData(Structure):
        'room_direction_data'

        _fields_ = \
            [('general_description', c_char_p),
             ('keyword', c_char_p),
             ('exit_info', c_long),
             ('key', obj_vnum),
             ('to_room', room_rnum)]

    DirectionData.P = POINTER(DirectionData)

RoomData.P = POINTER(RoomData)
RoomData._fields_ = \
    [('number', room_vnum),
     ('zone', zone_rnum),
     ('sector_type', c_int),
     ('name', c_char_p),
     ('description', c_char_p),
     ('ex_description', ExtraDescrData.P),
     ('dir_option', RoomData.DirectionData.P * RoomData.NUM_OF_DIRS),
     ('room_flags', c_int),
     ('affected_by', CharacterData.AffectedType.P),
     ('light', c_byte),
     ('func', SPECIAL_TYPE),
     ('contents', ItemData.P),
     ('people', CharacterData.P),
     ('py_handle', c_void_p)]

class TrustCommandInfo(Structure):
    'trust_cmd_info'

TrustCommandInfo.P = POINTER(TrustCommandInfo)
TrustCommandInfo._fields_ = \
    [('command', c_int),
     ('next', TrustCommandInfo.P)]

class AttackHitType(Structure):
    'attack_hit_type'

    _fields_ = \
        [('singular', c_char_p),
         ('plural', c_char_p)]

AttackHitType.P = POINTER(AttackHitType)

class AuctionData(Structure):
    'auction_data'

AuctionData.P = POINTER(AuctionData)
AuctionData._fields_ = \
    [('obj', ItemData.P),
     ('seller', CharacterData.P),
     ('bidder', CharacterData.P),
     ('min_bid', c_int),
     ('current_bid', c_int),
     ('timer', c_int)]

class BanListElement(Structure):
    'ban_list_element'

BanListElement.P = POINTER(BanListElement)
BanListElement._fields_ = \
    [('start_ip', c_long),
     ('end_ip', c_long),
     ('when', c_time_t),
     ('type', c_byte),
     ('by', c_byte * (MAX_NAME_LENGTH+1)),
     ('next', BanListElement.P)]

class BoardInfoType(Structure):
    'board_info_type'

    _fields_ = \
         [('vnum', obj_vnum),
          ('read_lvl', c_int),
          ('write_lvl', c_int),
          ('remove_lvl', c_int),
          ('filename', c_byte * 128),
          ('cleanup', c_int),
          ('rnum', obj_rnum)]

    class MessageInfo(Structure):
        'board_msginfo'

        _fields_ = \
             [('slot_num', c_int),
              ('heading', c_char_p),
              ('level', c_int),
              ('heading_len', c_int),
              ('message_len', c_int),
              ('birth', c_time_t)]

    MessageInfo.PP = POINTER(POINTER(MessageInfo))

BoardInfoType.P = POINTER(BoardInfoType)

class ClassInfoType(Structure):
    'class_info_type'

    _fields_ = \
        [('name', c_char_p),
         ('abbrev_color', c_char_p),
         ('abbrev_no_color', c_char_p),
         ('anti', c_long)]

ClassInfoType.P = POINTER(ClassInfoType)

class CommandInfo(Structure):
    'command_info'

    _fields_ = \
        [('command', c_char_p),
         ('minimum_position', c_byte),
         ('command_pointer', CFUNCTYPE(c_void, CharacterData.P, c_char_p, c_int, c_int)),
         ('minimum_level', c_short),
         ('subcmd', c_int)]

CommandInfo.P = POINTER(CommandInfo)

class FightMessageList(Structure):
    'message_list'

    class MessageType(Structure):
        'message_type'

        class Message(Structure):
            'msg_type'

            _fields_ = \
                [('attacker_msg', c_char_p),
                 ('victim_msg', c_char_p),
                 ('room_msg', c_char_p)]

    MessageType.P = POINTER(MessageType)
    MessageType._fields_ = \
        [('die_msg', MessageType.Message),
         ('miss_msg', MessageType.Message),
         ('hit_msg', MessageType.Message),
         ('god_msg', MessageType.Message),
         ('next', MessageType.P)]

FightMessageList.P = POINTER(FightMessageList)
FightMessageList._fields_ = \
    [('a_type', c_int),
     ('number_of_attacks', c_int),
     ('msg', FightMessageList.MessageType.P)]

class HelpEntry(Structure):
    'help_entry'

    _fields_= \
        [('keywords', c_char_p),
         ('text', c_char_p),
         ('minlevel', c_int),
         ('savefile', c_char_p)]

HelpEntry.P = POINTER(HelpEntry)

class HometownData(Structure):
    'hometown_data'

    _fields_= \
        [('name', c_char_p),
         ('start_room', c_int),
         ('low', c_int),
         ('high', c_int)]

HometownData.P = POINTER(HometownData)

class HouseControlRecord(Structure):
    'house_control_rec'
    MAX_GUESTS = 10

    _fields_= \
        [('vnum', room_vnum),
         ('atrium', room_vnum),
         ('exit_num', c_short),
         ('built_on', c_time_t),
         ('mode', c_int),
         ('owner', c_long),
         ('num_of_guests', c_int),
         ('guests', c_long * MAX_GUESTS),
         ('last_payment', c_time_t),
         ('max_item_save', c_long),
         ('spare1', c_long),
         ('spare2', c_long),
         ('spare3', c_long),
         ('spare4', c_long),
         ('spare5', c_long),
         ('spare6', c_long),
         ('spare7', c_long)]

HouseControlRecord.P = POINTER(HouseControlRecord)

class IndexData(Structure):
    'index_data'

    _fields_ = \
        [('vnum', VirtualNumber),
         ('number', c_int),
         ('func', SPECIAL_TYPE)]

IndexData.P = POINTER(IndexData)

class RaceInfoType(Structure):
    'race_info_type'

    _fields_ = \
        [('name', c_char_p),
         ('color_name', c_char_p),
         ('abbrev_color', c_char_p),
         ('abbrev_no_color', c_char_p),
         ('max_str', c_int),
         ('max_int', c_int),
         ('max_wis', c_int),
         ('max_dex', c_int),
         ('max_con', c_int),
         ('max_cha', c_int),
         ('min_remort_level', c_int),
         ('innates', c_int_p)]

RaceInfoType.P = POINTER(RaceInfoType)

class ResetQType(Structure):
    'reset_q_type'

    class Element(Structure):
        'reset_q_element'

    Element.P = POINTER(Element)
    Element._fields_ = \
        [('zone_to_reset', zone_rnum),
         ('next', POINTER(Element))]

ResetQType._fields_ = \
    [('head', ResetQType.Element.P),
     ('tail', ResetQType.Element.P)]

class ShopData(Structure):
    'shop_data'

    class BuyData(Structure):
        'shop_buy_data'

        _fields_ = \
            [('type', c_int),
             ('keywords', c_char_p)]

    BuyData.P = POINTER(BuyData)

    _fields_ = \
        [('zone', zone_vnum),
         ('vnum', room_vnum),
         ('producing', POINTER(obj_vnum)),
         ('profit_buy', c_float),
         ('profit_sell', c_float),
         ('type', BuyData.P),
         ('no_such_item1', c_char_p),
         ('no_such_item2', c_char_p),
         ('missing_cash1', c_char_p),
         ('missing_cash2', c_char_p),
         ('do_not_buy', c_char_p),
         ('message_buy', c_char_p),
         ('message_sell', c_char_p),
         ('temper1', c_int),
         ('bitvector', c_int),
         ('keeper', mob_rnum),
         ('with_who', c_int),
         ('in_room', POINTER(room_vnum)),
         ('open1', c_int),
         ('open2', c_int),
         ('close1', c_int),
         ('close2', c_int),
         ('bankAccount', c_int),
         ('goldOnHand', c_int),
         ('lastsort', c_int),
         ('func', SPECIAL_TYPE)]

ShopData.P = POINTER(ShopData)

class SocialListData(Structure):
    'social_list_data'

    class SocialData(Structure):
        'social_data'

    SocialData.P = POINTER(SocialData)
    SocialData._fields_ = \
        [('char_no_args', c_char_p),
         ('others_no_args', c_char_p),
         ('char_found', c_char_p),
         ('others_found', c_char_p),
         ('vict_found', c_char_p),
         ('not_found', c_char_p),
         ('char_auto', c_char_p),
         ('others_auto', c_char_p),
         ('next', SocialData.P)]

SocialListData.P = POINTER(SocialListData)
SocialListData._fields_ = \
    [('cmd', c_int),
     ('position', c_int),
     ('hide', c_int),
     ('socials', SocialListData.SocialData.P),
     ('next', SocialListData.P)]

class TimeInfoData(Structure):
    'time_info_data'

    _fields_ = \
        [('hours', c_int),
         ('day', c_int),
         ('month', c_int),
         ('year', c_short)]

TimeInfoData.P = POINTER(TimeInfoData)

class TitleType(Structure):
    'title_type'

    _fields_ = \
        [('title_m', c_char_p),
         ('title_f', c_char_p),
         ('exp', c_int)]

TitleType.P  = POINTER(TitleType  )
TitleType.PP = POINTER(TitleType.P)

class WeatherData(Structure):
    'weather_data'

    _fields_ = \
        [('pressure', c_int),
         ('change', c_int),
         ('sky', c_int),
         ('sunlight', c_int)]

WeatherData.P = POINTER(WeatherData)

class OlcIndexData(Structure):
    'olc_index_data'

    class OlcListData(Structure):
        'olc_list_data'

    OlcListData.P = POINTER(OlcListData)
    OlcListData._fields_ = \
        [('idnum', c_long),
         ('next', OlcListData.P)]

OlcIndexData.P = POINTER(OlcIndexData)
OlcIndexData._fields_ = [('list', OlcIndexData.OlcListData.P)]

class PlayerIndexElement(Structure):
    'player_index_element'

    _fields_ = \
        [('lvl', c_int),
         ('rname', c_char_p),
         ('name', c_char_p),
         ('id', c_long),
         ('commands', TrustCommandInfo.P)]

PlayerIndexElement.P = POINTER(PlayerIndexElement)

class MailData(Structure):
    'mail_data'

    class Recipient(Structure):
        'mail_to_data'

MailData.Recipient.P = POINTER(MailData.Recipient)
MailData.Recipient._fields_ = \
    [('idnum', c_long),
     ('next', POINTER(MailData.Recipient))]

MailData.P = POINTER(MailData)
MailData._fields_ = \
    [('from', c_char_p),
     ('subject', c_char_p),
     ('body', c_char_p),
     ('to', MailData.Recipient.P),
     ('next', MailData.P)]

class ClanInfoType(Structure):
    'clan_info_type'

    class ClanRankType(Structure):
        'clan_rank_type'
        NUM_CLAN_RANKS = 10

        _fields_ = \
            [('male', c_char_p),
             ('female', c_char_p)]

ClanInfoType.P = POINTER(ClanInfoType)
ClanInfoType._fields_= \
    [('vnum', c_int),
     ('alias', c_char_p),
     ('name', c_char_p),
     ('color_name', c_char_p),
     ('filename', c_char_p),
     ('donate_room', c_int),
     ('entrance_room', c_int),
     ('recall_room', c_int),
     ('range_low_room', c_int),
     ('range_high_room', c_int),
     ('store_object', c_int),
     ('last_updated', c_time_t),
     ('updated_by', c_long),
     ('ranks', ClanInfoType.ClanRankType * ClanInfoType.ClanRankType.NUM_CLAN_RANKS),
     ('next', ClanInfoType.P)]

class ZoneData(Structure):
    'zone_data'

    class ZCmdData(Structure):
        'reset_com'

        _fields_ = \
            [('command', c_byte),
             ('if_flag', c_bool),
             ('arg1', c_int),
             ('arg2', c_int),
             ('arg3', c_int),
             ('line', c_int)]

ZoneData.P = POINTER(ZoneData)
ZoneData._fields_ = \
    [('name', c_char_p),
     ('lifespan', c_int),
     ('age', c_int),
     ('bot', room_vnum),
     ('top', room_vnum),
     ('zone_flags', c_long),
     ('reset_mode', c_int),
     ('number', zone_vnum),
     ('cmd', POINTER(ZoneData.ZCmdData)),
     ('continent', c_int),
     ('func', SPECIAL_TYPE),
     ('py_handle', c_void_p)]

class RemortEdit(Structure):
    'RemortEdit'
    # C++ class.  Prototyped here for compatability.
    # The field data here is expressed arbitrarily
    # because the ctypes FFI code requires a sizable
    # structure on certain platforms (I think).

RemortEdit.P = POINTER(RemortEdit)

class OLCZCmdData(Structure):
    'zcmd_data'
    # Note that this is different from reset_com,
    # which is defined as a ctype in the ZoneData class.

    _fields_ = \
        [('command', c_byte),
         ('if_flag', c_bool),
         ('arg1', c_int),
         ('arg2', c_int),
         ('arg3', c_int)]

DescriptorData._fields_ = \
    [('remort', RemortEdit.P),
     ('descriptor', c_socket_t),
     ('ident_sock', c_socket_t),
     ('peer_port', c_short),
     ('network_mode', c_int),
     ('host', c_byte * (DescriptorData.HOST_LENGTH+1)),
     ('ip', c_long),
     ('port', c_short),
     ('login', c_time_t),
     ('last_host', c_byte * (DescriptorData.HOST_LENGTH+1)),
     ('bad_pws', c_byte),
     ('idle_tics', c_byte),
     ('connected', c_int),
     ('desc_num', c_int),
     ('login_time', c_time_t),
     ('command_time', c_time_t),
     ('showstr_head', c_char_p),
     ('showstr_vector', c_char_pp),
     ('showstr_count', c_int),
     ('showstr_page', c_int),
     ('showstr_point', c_char_p),
     ('showstr_start', c_char_p),
     ('str', c_char_pp),
     ('email', c_char_p),
     ('max_str', c_size_t),
     ('mail_to', c_long),
     ('mail_mode', c_int),
     ('mail', MailData.P),
     ('report', c_char_p),
     ('has_prompt', c_int),
     ('color_mode', c_int),
     ('medit_mode', c_int),
     ('oedit_mode', c_int),
     ('sedit_mode', c_int),
     ('zedit_mode', c_int),
     ('zedit_page', c_int),
     ('zedit_index', c_int),
     ('hedit_mode', c_int),
     ('cedit_mode', c_int),
     ('cedit_clan', ClanInfoType.P),
     ('olc_save_type', c_int),
     ('vt_bar_on', c_int),
     ('saved', c_int),
     ('affect_index', c_int),
     ('edit_vnum', c_int),
     ('edit_room', c_int),
     ('edit_zone', c_int),
     ('edit_clan', c_int),
     ('help_entry', c_int),
     ('redit_dir', c_int),
     ('redit_mode', c_int),
     ('exdesc', ExtraDescrData.P),
     ('prev_exdesc', ExtraDescrData.P),
     ('edit_obj', ItemData.P),
     ('edit_mob', CharacterData.P),
     ('edit_shop', ShopData.P),
     ('cmd', OLCZCmdData),
     ('olc_string', c_byte * MAX_INPUT_LENGTH),
     ('olc_int', c_int),
     ('gen_spec_act', POINTER(CharacterData.NpcSpecialData.NpcActionList)),
     ('inbuf', c_byte * MAX_RAW_INPUT_LENGTH),
     ('last_input', c_byte * MAX_INPUT_LENGTH),
     ('small_outbuf', c_byte * SMALL_BUFSIZE),
     ('history', c_char_pp),
     ('history_pos', c_int),
     ('bufptr', c_size_t),
     ('bufspace', c_size_t),
     ('large_outbuf', POINTER(DescriptorData.TxtBlock)),
     ('input', DescriptorData.TxtQueue),
     ('character', CharacterData.P),
     ('original', CharacterData.P),
     ('snooping', DescriptorData.P),
     ('snoop_by', DescriptorData.P),
     ('next', DescriptorData.P),
     ('bandwidth_in', c_long),
     ('bandwidth_out', c_long),
     ('py_handle', c_void_p),
     ('telnetOption', CFUNCTYPE(c_void, DescriptorData.P, c_byte, c_byte, c_char_p))]

class EventOperations(Structure):
    'EventOperations'
    _fields_ = \
        [('boot_start', CFUNCTYPE(c_void)),
         ('boot_complete', CFUNCTYPE(TimeValue.P)),
         ('world_reset_start', CFUNCTYPE(c_void)),
         ('world_reset_complete', CFUNCTYPE(c_void)),
         ('heartbeat_pulse', CFUNCTYPE(c_char_p, c_int)),
         ('heartbeat_timeout', CFUNCTYPE(TimeValue.P)),
         ('do_python', CFUNCTYPE(CharacterData.P, c_char_p, c_int, c_int))]
        # Unfinished

EventOperations.P = POINTER(EventOperations)

# Todo: reconfigure getvalues to use the autofunc.
class MUDOperations(Structure):
    'MUDOperations'

    _fields_ = \
       [('BankAccess'                  , CFUNCTYPE(c_int, c_int, c_int, c_int)),
        ('CAP'                         , CFUNCTYPE(c_char_p, c_char_p)),
        ('Crash_load'                  , CFUNCTYPE(c_int, CharacterData.P)),
        ('Crash_rentsave'              , CFUNCTYPE(c_int, CharacterData.P, c_int)),
        ('Crash_save_all'              , CFUNCTYPE(c_int)),
        ('FindBanTypeDesc'             , CFUNCTYPE(c_int, c_int)),
        ('House_can_enter'             , CFUNCTYPE(c_int, CharacterData.P, c_int)),
        ('House_crashsave'             , CFUNCTYPE(c_int, c_int)),
        ('House_delete_file'           , CFUNCTYPE(c_int, c_int)),
        ('House_get_filename'          , CFUNCTYPE(c_int, c_int, c_char_p, c_ulong)),
        ('House_load'                  , CFUNCTYPE(c_int, c_int)),
        ('House_save_all'              , CFUNCTYPE(c_int)),
        ('InstallStuphEventBridge'     , CFUNCTYPE(c_void, EventOperations)),
        ('IsBanned'                    , CFUNCTYPE(c_int, c_ulong)),
        ('IsBannedByHost'              , CFUNCTYPE(c_int, c_char_p)),
        ('LoadBanned'                  , CFUNCTYPE(c_int)),
        ('Obj_from_store_to'           , AUTOFUNCTYPE(ItemData.P, ItemRentData.P, c_int_p)),
        ('Obj_to_store_temp'           , CFUNCTYPE(c_int, ItemData.P, ItemRentData.P, c_int)),
        ('PyAcquireControl'            , CFUNCTYPE(c_int)),
        ('PyExecuteFile'               , CFUNCTYPE(c_int, c_char_p)),
        ('PyReleaseControl'            , CFUNCTYPE(c_int)),
        ('Py_IsInteractive'            , CFUNCTYPE(c_int)),
        ('StuphItemPrototype_Created'  , CFUNCTYPE(c_int, c_int)),
        ('StuphMobile_Extraction'      , CFUNCTYPE(c_int, CharacterData.P)),
        ('StuphPeer_QuitGame'          , CFUNCTYPE(c_int, DescriptorData.P)),
        ('StuphWorld_ResetComplete'    , CFUNCTYPE(c_int)),
        ('StuphWorld_ResetStart'       , CFUNCTYPE(c_int)),
        ('Valid_Name'                  , CFUNCTYPE(c_int, c_char_p)),
        ('add_to_dns_cache'            , CFUNCTYPE(c_int, c_ulong, c_char_p)),
        ('add_to_olc_index'            , CFUNCTYPE(c_int, c_int, c_long)),
        ('announce'                    , CFUNCTYPE(c_void, CharacterData.P, c_bool, c_char_p, c_char_p)),
        ('apply_ac'                    , CFUNCTYPE(c_int, CharacterData.P, c_int)),
        ('asciiflag_conv'              , CFUNCTYPE(c_int, c_char_p)),
        ('auction_update'              , CFUNCTYPE(c_int)),
        ('backup'                      , CFUNCTYPE(c_int)),
        ('backup_filename'             , CFUNCTYPE(c_char_p)),
        ('backup_freq'                 , CFUNCTYPE(c_int)),
        ('bad_help_entry'              , CFUNCTYPE(c_int, c_int)),
        ('basic_mud_log'               , CFUNCTYPE(c_void, c_char_p, c_char_p)),
        ('blank_profanity'             , CFUNCTYPE(c_int, c_char_p, c_char_pp)),
        ('bufcat'                      , CFUNCTYPE(c_int, c_long, c_char_p)),
        ('bufcpy'                      , CFUNCTYPE(c_int, c_long, c_char_p)),
        ('buflen'                      , CFUNCTYPE(c_int, c_long)),
        ('buildWorldPathDEBUG'         , CFUNCTYPE(c_char_p, c_char_p, c_int, c_char_p, c_char_p)),
        ('build_olc_index'             , CFUNCTYPE(c_int)),
        ('can_create_zone'             , CFUNCTYPE(c_int, c_int)),
        ('change_alignment'            , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('char_from_room'              , CFUNCTYPE(c_int, CharacterData.P)),
        ('char_to_room'                , CFUNCTYPE(c_int, CharacterData.P, c_int)),
        ('char_to_store'               , CFUNCTYPE(c_void, CharacterData.P, CharacterData.FileU.P)),
        ('check_autoassist'            , CFUNCTYPE(c_int, CharacterData.P)),
        ('check_idle_passwords'        , CFUNCTYPE(c_int)),
        ('check_subdue'                , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('cir_record_usage'            , CFUNCTYPE(c_int)),
        ('clear_char'                  , CFUNCTYPE(c_void, CharacterData.P)),
        ('command_interpreter'         , CFUNCTYPE(c_void, CharacterData.P, c_char_p)),
        ('compute_armor_class'         , CFUNCTYPE(c_int, CharacterData.P)),
        ('compute_thaco'               , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('count_hash_records'          , CFUNCTYPE(c_int, FILE_P)),
        ('create_char'                 , AUTOFUNCTYPE(CharacterData.P)),
        ('create_money'                , AUTOFUNCTYPE(ItemData.P, c_int)),
        ('damage'                      , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P, c_int, c_int)),
        ('death_cry'                   , CFUNCTYPE(c_int, CharacterData.P)),
        ('die'                         , CFUNCTYPE(c_int, CharacterData.P)),
        ('dns_cache_lookup'            , CFUNCTYPE(c_char_p, c_ulong)),
        ('do_action'                   , CFUNCTYPE(c_void, CharacterData.P, c_char_p, c_int, c_int)),
        ('do_auto_exits'               , CFUNCTYPE(c_int, CharacterData.P)),
        ('do_insult'                   , CFUNCTYPE(c_void, CharacterData.P, c_char_p, c_int, c_int)),
        ('do_start'                    , CFUNCTYPE(c_int, CharacterData.P, c_int)),
        ('drop_norents'                , CFUNCTYPE(c_int, CharacterData.P, ItemData.P)),
        ('echo_off'                    , CFUNCTYPE(c_int, DescriptorData.P)),
        ('echo_on'                     , CFUNCTYPE(c_int, DescriptorData.P)),
        ('equip_char'                  , AUTOFUNCTYPE(ItemData.P, CharacterData.P, ItemData.P, c_int, c_int)),
        ('extract_char'                , CFUNCTYPE(c_int, CharacterData.P)),
        ('extract_obj'                 , CFUNCTYPE(c_int, ItemData.P)),
        ('extract_pending_chars'       , CFUNCTYPE(c_void)),
        ('fcopy'                       , CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('find_all_dots'               , CFUNCTYPE(c_int, c_char_p)),
        ('find_exdesc'                 , CFUNCTYPE(c_char_p, c_char_p, ExtraDescrData.P)),
        ('find_first_step'             , CFUNCTYPE(c_int, c_int, c_int, c_int_p)),
        ('find_help_entry'             , CFUNCTYPE(c_int, c_char_p, c_int)),
        ('find_house'                  , CFUNCTYPE(c_int, c_int)),
        ('find_room_insert_position'   , CFUNCTYPE(c_int, c_int)),
        ('find_zone'                   , CFUNCTYPE(c_int, c_int)),
        ('find_zone_insert_position'   , CFUNCTYPE(c_int, c_int)),
        ('flush_queues'                , CFUNCTYPE(c_int, DescriptorData.P)),
        ('fname'                       , CFUNCTYPE(c_char_p, c_char_p)),
        ('fread_string'                , CFUNCTYPE(c_char_p, FILE_P, c_char_p)),
        ('free_char'                   , CFUNCTYPE(c_void, CharacterData.P)),
        ('freebuf'                     , CFUNCTYPE(c_int, c_long)),
        ('generic_find'                , CFUNCTYPE(c_int, c_char_p, c_ulong, CharacterData.P, CharacterData.PP, ItemData.PP)),
        ('gettimeofday'                , CFUNCTYPE(c_int, c_timeval.P, c_void_p)),
        ('get_bufstr'                  , CFUNCTYPE(c_char_p, c_long)),
        ('get_char_num'                , AUTOFUNCTYPE(CharacterData.P, c_int)),
        ('get_char_room_vis'           , AUTOFUNCTYPE(CharacterData.P, CharacterData.P, c_char_p, c_int_p)),
        ('get_char_world_vis'          , AUTOFUNCTYPE(CharacterData.P, CharacterData.P, c_char_p, c_int_p)),
        ('get_commands_by_id'          , AUTOFUNCTYPE(TrustCommandInfo.P, c_long)),
        ('get_filename'                , CFUNCTYPE(c_int, c_char_p, c_ulong, c_int, c_char_p)),
        ('get_from_q'                  , CFUNCTYPE(c_int, DescriptorData.TxtQueue.P, c_char_p, c_int_p)),
        ('get_hometown_by_room'        , CFUNCTYPE(c_int, c_int)),
        ('get_hostname'                , CFUNCTYPE(c_char_p, c_ulong)),
        ('get_id_by_name'              , CFUNCTYPE(c_long, c_char_p)),
        ('get_last_backup'             , CFUNCTYPE(c_int)),
        ('get_level_by_id'             , CFUNCTYPE(c_int, c_long)),
        ('get_line'                    , CFUNCTYPE(c_int, FILE_P, c_char_p)),
        ('get_mail_count'              , CFUNCTYPE(c_int, c_long, c_int)),
        ('get_name_by_id'              , CFUNCTYPE(c_char_p, c_long)),
        ('get_number'                  , CFUNCTYPE(c_int, c_char_pp)),
        ('get_obj_in_list_vis'         , AUTOFUNCTYPE(ItemData.P, CharacterData.P, c_char_p, c_int_p, ItemData.P)),
        ('get_obj_num'                 , AUTOFUNCTYPE(ItemData.P, c_int)),
        ('get_player_vis'              , AUTOFUNCTYPE(CharacterData.P, CharacterData.P, c_char_p, c_int_p, c_int, c_int)),
        ('get_ptable_by_id'            , CFUNCTYPE(c_long, c_int)),
        ('get_ptable_by_name'          , CFUNCTYPE(c_long, c_char_p)),
        ('getvalue_AGE_NOTICE'         , CFUNCTYPE(c_char_p)),
        ('getvalue_DFLT_PORT'          , CFUNCTYPE(c_int)),
        ('getvalue_EMAIL_NOTICE'       , CFUNCTYPE(c_char_p)),
        ('getvalue_MENU'               , CFUNCTYPE(c_char_p)),
        ('getvalue_PEACEFUL_MSG'       , CFUNCTYPE(c_char_p)),
        ('getvalue_START_MESSG'        , CFUNCTYPE(c_char_p)),
        ('getvalue_WELC_MESSG'         , CFUNCTYPE(c_char_p)),
        ('getvalue_WIZLOCK_MSG'        , CFUNCTYPE(c_char_p)),
        ('getvalue_player_bits'        , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_preference_bits'    , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_preference_bits2'   , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_action_bits'        , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_room_bits'          , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_extra_bits'         , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_wear_bits'          , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_affected_bits'      , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_all_start_equip'    , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_apply_types'        , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_attack_hit_text'    , AUTOFUNCTYPE(AttackHitType.P)),
        ('getvalue_attack_mode_names'  , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_attack_mode_order'  , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_auction'            , AUTOFUNCTYPE(AuctionData.P)),
        ('getvalue_auto_save'          , CFUNCTYPE(c_int)),
        ('getvalue_autosave_time'      , CFUNCTYPE(c_int)),
        ('getvalue_background'         , CFUNCTYPE(c_char_p)),
        ('getvalue_backups_disabled'   , CFUNCTYPE(c_int)),
        ('getvalue_ban_list'           , AUTOFUNCTYPE(BanListElement.P)),
        ('getvalue_board_info'         , AUTOFUNCTYPE(BoardInfoType.P)),
        ('getvalue_boot_time'          , CFUNCTYPE(c_int)),
        ('getvalue_buf_largecount'     , CFUNCTYPE(c_int)),
        ('getvalue_buf_overflows'      , CFUNCTYPE(c_int)),
        ('getvalue_buf_switches'       , CFUNCTYPE(c_int)),
        ('getvalue_bufpool'            , AUTOFUNCTYPE(DescriptorData.TxtBlock.P)),
        ('getvalue_bugs'               , CFUNCTYPE(c_char_p)),
        ('getvalue_channel_buffer'     , AUTOFUNCTYPE(LastBuffer.P)),
        ('getvalue_channel_buffer_names', AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_character_list'     , AUTOFUNCTYPE(CharacterData.P)),
        ('getvalue_circle_reboot'      , CFUNCTYPE(c_int)),
        ('getvalue_circle_restrict'    , CFUNCTYPE(c_int)),
        ('getvalue_circle_shutdown'    , CFUNCTYPE(c_int)),
        ('getvalue_circlemud_version'  , CFUNCTYPE(c_char_p)),
        ('getvalue_clan_help'          , CFUNCTYPE(c_char_p)),
        ('getvalue_class_count'        , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_class_info'         , AUTOFUNCTYPE(ClassInfoType.P)),
        ('getvalue_class_names'        , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_cmd_info'           , AUTOFUNCTYPE(CommandInfo.P)),
        ('getvalue_cmd_sort_info'      , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_combat_list'        , AUTOFUNCTYPE(CharacterData.P)),
        ('getvalue_connected_types'    , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_crash_file_timeout' , CFUNCTYPE(c_int)),
        ('getvalue_credits'            , CFUNCTYPE(c_char_p)),
        ('getvalue_dirs'               , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_donation_room_1'    , CFUNCTYPE(c_int)),
        ('getvalue_donation_room_2'    , CFUNCTYPE(c_int)),
        ('getvalue_donation_room_3'    , CFUNCTYPE(c_int)),
        ('getvalue_dts_are_dumps'      , CFUNCTYPE(c_int)),
        ('getvalue_efficiency'         , CFUNCTYPE(c_int)),
        ('getvalue_emergency_unban'    , CFUNCTYPE(c_int)),
        ('getvalue_exit_bits'          , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_fight_messages'     , AUTOFUNCTYPE(FightMessageList.P)),
        ('getvalue_free_rent'          , CFUNCTYPE(c_int)),
        ('getvalue_frozen_start_room'  , CFUNCTYPE(c_int)),
        ('getvalue_genders'            , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_generic_trash_vnum' , CFUNCTYPE(c_int)),
        ('getvalue_handbook'           , CFUNCTYPE(c_char_p)),
        ('getvalue_help'               , CFUNCTYPE(c_char_p)),
        ('getvalue_help_rename'        , CFUNCTYPE(c_char_p)),
        ('getvalue_help_table'         , AUTOFUNCTYPE(HelpEntry.P)),
        ('getvalue_holler_move_cost'   , CFUNCTYPE(c_int)),
        ('getvalue_hometown_count'     , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_hometowns'          , AUTOFUNCTYPE(HometownData.P)),
        ('getvalue_house_control'      , AUTOFUNCTYPE(HouseControlRecord.P)),
        ('getvalue_ideas'              , CFUNCTYPE(c_char_p)),
        ('getvalue_idle_max_level'     , CFUNCTYPE(c_int)),
        ('getvalue_idle_rent_time'     , CFUNCTYPE(c_int)),
        ('getvalue_idle_timeout_check' , CFUNCTYPE(c_int)),
        ('getvalue_idle_void'          , CFUNCTYPE(c_int)),
        ('getvalue_immlist'            , CFUNCTYPE(c_char_p)),
        ('getvalue_immort_level_ok'    , CFUNCTYPE(c_int)),
        ('getvalue_immort_start_room'  , CFUNCTYPE(c_int)),
        ('getvalue_imotd'              , CFUNCTYPE(c_char_p)),
        ('getvalue_improve_menu'       , CFUNCTYPE(c_char_p)),
        ('getvalue_info'               , CFUNCTYPE(c_char_p)),
        ('getvalue_item_types'         , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_level_can_shout'    , CFUNCTYPE(c_int)),
        ('getvalue_loadIntoObjectIndex', CFUNCTYPE(obj_rnum)),
        ('getvalue_loadIntoZoneIndex'  , CFUNCTYPE(zone_rnum)),
        ('getvalue_load_into_inventory', CFUNCTYPE(c_int)),
        ('getvalue_log_last_cmd'       , CFUNCTYPE(c_int)),
        ('getvalue_max_aliases'        , CFUNCTYPE(c_int)),
        ('getvalue_max_bad_pws'        , CFUNCTYPE(c_int)),
        ('getvalue_max_breath_time'    , CFUNCTYPE(c_int)),
        ('getvalue_max_exp_gain'       , CFUNCTYPE(c_int)),
        ('getvalue_max_exp_loss'       , CFUNCTYPE(c_int)),
        ('getvalue_max_filesize'       , CFUNCTYPE(c_int)),
        ('getvalue_max_gold_decay_timer', CFUNCTYPE(c_int)),
        ('getvalue_max_house_save'     , CFUNCTYPE(c_int)),
        ('getvalue_max_key_time'       , CFUNCTYPE(c_int)),
        ('getvalue_max_npc_corpse_time', CFUNCTYPE(c_int)),
        ('getvalue_max_obj_save'       , CFUNCTYPE(c_int)),
        ('getvalue_max_object_decay_timer', CFUNCTYPE(c_int)),
        ('getvalue_max_pc_corpse_time' , CFUNCTYPE(c_int)),
        ('getvalue_max_pfile_boots'    , CFUNCTYPE(c_int)),
        ('getvalue_max_players'        , CFUNCTYPE(c_int)),
        ('getvalue_max_playing'        , CFUNCTYPE(c_int)),
        ('getvalue_max_spell_attacks'  , CFUNCTYPE(c_int)),
        ('getvalue_min_passwd_length'  , CFUNCTYPE(c_int)),
        ('getvalue_min_rent_cost'      , CFUNCTYPE(c_int)),
        ('getvalue_min_wizlist_lev'    , CFUNCTYPE(c_int)),
        ('getvalue_mini_mud'           , CFUNCTYPE(c_int)),
        ('getvalue_mob_index'          , AUTOFUNCTYPE(IndexData.P)),
        ('getvalue_mob_proto'          , AUTOFUNCTYPE(CharacterData.P)),
        ('getvalue_mortal_start_room'  , CFUNCTYPE(room_vnum)),
        ('getvalue_motd'               , CFUNCTYPE(c_char_p)),
        ('getvalue_msg_index'          , AUTOFUNCTYPE(BoardInfoType.MessageInfo.PP)),
        ('getvalue_msg_storage'        , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_mudlog_types'       , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_nameserver_is_slow' , CFUNCTYPE(c_int)),
        ('getvalue_newbie_ban'         , CFUNCTYPE(c_char_p)),
        ('getvalue_newbie_level'       , CFUNCTYPE(c_int)),
        ('getvalue_newbie_start_room'  , CFUNCTYPE(c_int)),
        ('getvalue_news'               , CFUNCTYPE(c_char_p)),
        ('getvalue_no_rent_check'      , CFUNCTYPE(c_int)),
        ('getvalue_no_specials'        , CFUNCTYPE(c_int)),
        ('getvalue_num_invalid'        , CFUNCTYPE(c_int)),
        ('getvalue_num_of_msgs'        , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_obj_index'          , AUTOFUNCTYPE(IndexData.P)),
        ('getvalue_obj_proto'          , AUTOFUNCTYPE(ItemData.P)),
        ('getvalue_object_list'        , AUTOFUNCTYPE(ItemData.P)),
        ('getvalue_olc_index'          , AUTOFUNCTYPE(OlcIndexData.P)),
        ('getvalue_pk_allowed'         , CFUNCTYPE(c_int)),
        ('getvalue_pkok_in_peaceful_rooms', CFUNCTYPE(c_int)),
        ('getvalue_player_fl'          , CFUNCTYPE(FILE_P)),
        ('getvalue_player_table'       , AUTOFUNCTYPE(PlayerIndexElement.P)),
        ('getvalue_policies'           , CFUNCTYPE(c_char_p)),
        ('getvalue_port'               , CFUNCTYPE(c_int)),
        ('getvalue_position_types'     , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_pt_allowed'         , CFUNCTYPE(c_int)),
        ('getvalue_r_frozen_start_room', CFUNCTYPE(room_rnum)),
        ('getvalue_r_immort_start_room', CFUNCTYPE(room_rnum)),
        ('getvalue_r_mortal_start_room', CFUNCTYPE(room_rnum)),
        ('getvalue_r_newbie_start_room', CFUNCTYPE(room_rnum)),
        ('getvalue_race_desc'          , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_race_info'          , AUTOFUNCTYPE(RaceInfoType.P)),
        ('getvalue_race_names'         , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_rent_file_timeout'  , CFUNCTYPE(c_int)),
        ('getvalue_reread_wizlist'     , CFUNCTYPE(c_int)),
        ('getvalue_reset_q'            , AUTOFUNCTYPE(ResetQType)),
        ('getvalue_scheck'             , CFUNCTYPE(c_int)),
        ('getvalue_sector_types'       , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_select_ban'         , CFUNCTYPE(c_char_p)),
        ('getvalue_shop_index'         , AUTOFUNCTYPE(ShopData.P)),
        ('getvalue_shutdown_by'        , CFUNCTYPE(c_long)),
        ('getvalue_shutdown_mode'      , CFUNCTYPE(c_int)),
        ('getvalue_shutdown_modes'     , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_shutdown_timer'     , CFUNCTYPE(c_int)),
        ('getvalue_siteok_everyone'    , CFUNCTYPE(c_int)),
        ('getvalue_sky_types'          , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_social_list'        , AUTOFUNCTYPE(SocialListData.P)),
        ('getvalue_start_equip'        , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_sun_types'          , AUTOFUNCTYPE(c_char_pp)),
        ('getvalue_tics'               , CFUNCTYPE(c_int)),
        ('getvalue_time_info'          , AUTOFUNCTYPE(TimeInfoData)),
        ('getvalue_title_screen'       , CFUNCTYPE(c_char_p)),
        ('getvalue_titles'             , AUTOFUNCTYPE(TitleType.PP)),
        ('getvalue_top_of_mobt'        , CFUNCTYPE(c_int)),
        ('getvalue_top_of_objt'        , CFUNCTYPE(c_int)),
        ('getvalue_top_of_p_table'     , CFUNCTYPE(c_int)),
        ('getvalue_top_of_world'       , CFUNCTYPE(c_int)),
        ('getvalue_top_of_zone_table'  , CFUNCTYPE(c_int)),
        ('getvalue_top_shop'           , CFUNCTYPE(c_int)),
        ('getvalue_total_ban'          , CFUNCTYPE(c_char_p)),
        ('getvalue_track_through_doors', CFUNCTYPE(c_int)),
        ('getvalue_tunnel_size'        , CFUNCTYPE(c_int)),
        ('getvalue_typos'              , CFUNCTYPE(c_char_p)),
        ('getvalue_use_autowiz'        , CFUNCTYPE(c_int)),
        ('getvalue_use_dns_cache'      , CFUNCTYPE(c_int)),
        ('getvalue_usec_spent_per_sec' , CFUNCTYPE(c_int)),
        ('getvalue_wear_sort_pos'      , AUTOFUNCTYPE(c_int_p)),
        ('getvalue_weather_info'       , AUTOFUNCTYPE(WeatherData.P)),
        ('getvalue_wizlist'            , CFUNCTYPE(c_char_p)),
        ('getvalue_wiznames'           , CFUNCTYPE(c_int)),
        ('getvalue_world'              , AUTOFUNCTYPE(RoomData.P)),
        ('getvalue_zone_table'         , AUTOFUNCTYPE(ZoneData.P)),
        ('hit'                         , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P, c_int)),
        ('idle_dns_cache'              , CFUNCTYPE(c_int)),
        ('initworld'                   , CFUNCTYPE(c_int)),
        ('init_char'                   , CFUNCTYPE(c_void, CharacterData.P)),
        ('internalMakePrompt'          , CFUNCTYPE(c_char_p, DescriptorData.P)),
        ('invalid_align'               , CFUNCTYPE(c_int, CharacterData.P, ItemData.P)),
        ('is_empty'                    , CFUNCTYPE(c_int, c_int)),
        ('is_speedwalk_string'         , CFUNCTYPE(c_int, c_char_p)),
        ('isname'                      , CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('lightning'                   , CFUNCTYPE(c_int, CharacterData.P)),
        ('list_char_to_char'           , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('list_obj_to_char'            , CFUNCTYPE(c_int, ItemData.P, CharacterData.P, c_int, c_int)),
        ('load_char'                   , CFUNCTYPE(c_int, c_char_p, CharacterData.FileU.P)),
        ('load_dns_cache'              , CFUNCTYPE(c_int)),
        ('load_zones'                  , CFUNCTYPE(c_int, FILE_P, c_char_p)),
        ('lockMUD'                     , CFUNCTYPE(c_int, c_char_p)),
        ('log_zone_error'              , CFUNCTYPE(c_int, c_int, c_int, c_char_p)),
        ('mainDescriptorList'          , AUTOFUNCTYPE(DescriptorData.P)),
        ('make_corpse'                 , CFUNCTYPE(c_int, CharacterData.P)),
        ('mobiles_same'                , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('money_desc'                  , CFUNCTYPE(c_char_p, c_int)),
        ('mudlog'                      , CFUNCTYPE(c_void, c_int, c_int, c_int, c_char_p, c_char_p)),
        ('newbuf'                      , CFUNCTYPE(c_int)),
        ('num_of_zcmds'                , CFUNCTYPE(c_int, c_int)),
        ('obj_from_char'               , CFUNCTYPE(c_int, ItemData.P)),
        ('obj_from_obj'                , CFUNCTYPE(c_int, ItemData.P)),
        ('obj_from_room'               , CFUNCTYPE(c_int, ItemData.P)),
        ('obj_to_char'                 , CFUNCTYPE(c_int, ItemData.P, CharacterData.P)),
        ('obj_to_obj'                  , CFUNCTYPE(c_int, ItemData.P, ItemData.P)),
        ('obj_to_room'                 , CFUNCTYPE(c_int, ItemData.P, c_int)),
        ('open_logfile'                , CFUNCTYPE(c_int, c_char_p, FILE_P)),
        ('page_string'                 , CFUNCTYPE(c_int, DescriptorData.P, c_char_p, c_int)),
        ('parse_object'                , CFUNCTYPE(c_char_p, FILE_P, c_int)),
        ('perform_attack'              , CFUNCTYPE(c_int, CharacterData.P, c_int, c_int)),
        ('perform_backup'              , CFUNCTYPE(c_int)),
        ('perform_complex_alias'       , CFUNCTYPE(c_int, DescriptorData.TxtQueue.P, c_char_p, CharacterData.PlayerSpecialData.AliasData.P)),
        ('perform_subdue'              , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('power_rating'                , CFUNCTYPE(c_int, CharacterData.P)),
        ('process_speedwalk'           , CFUNCTYPE(c_int, DescriptorData.TxtQueue.P, c_char_p)),
        ('python_mobile'               , CFUNCTYPE(c_int, CharacterData.P, c_void_p, c_int, c_char_p)),
        ('python_object'               , CFUNCTYPE(c_int, CharacterData.P, c_void_p, c_int, c_char_p)),
        ('python_room'                 , CFUNCTYPE(c_int, CharacterData.P, c_void_p, c_int, c_char_p)),
        ('raw_kill'                    , CFUNCTYPE(c_int, CharacterData.P)),
        ('read_aliases'                , CFUNCTYPE(c_int, CharacterData.P)),
        ('read_config'                 , CFUNCTYPE(c_int, c_char_p)),
        ('read_last_backup'            , CFUNCTYPE(c_int)),
        ('read_mobile'                 , AUTOFUNCTYPE(CharacterData.P, c_int, c_int)),
        ('read_object'                 , AUTOFUNCTYPE(ItemData.P, c_int, c_int)),
        ('real_mobile'                 , CFUNCTYPE(mob_rnum, c_int)),
        ('real_object'                 , CFUNCTYPE(obj_rnum, c_int)),
        ('real_room'                   , CFUNCTYPE(room_rnum, c_int)),
        ('real_zone'                   , CFUNCTYPE(zone_rnum, c_int)),
        ('reboot_wizlists'             , CFUNCTYPE(c_int)),
        ('remove_from_olc_index'       , CFUNCTYPE(c_int, c_int, c_long)),
        ('report_dns'                  , CFUNCTYPE(c_int, c_char_p)),
        ('reset_time'                  , CFUNCTYPE(c_int)),
        ('reset_zone'                  , CFUNCTYPE(c_int, c_int)),
        ('saveZone'                    , CFUNCTYPE(c_int, c_int)),
        ('save_char'                   , CFUNCTYPE(c_void, CharacterData.P)),
        ('save_dns_cache'              , CFUNCTYPE(c_int)),
        ('save_load_rooms'             , CFUNCTYPE(c_int)),
        ('save_mud_time'               , CFUNCTYPE(c_void, TimeInfoData.P)),
        ('save_player_file'            , CFUNCTYPE(c_int, CharacterData.P, c_char_p, c_char_p)),
        ('saving_throws'               , CFUNCTYPE(c_int, c_int, c_int, c_int)),
        ('search_block'                , CFUNCTYPE(c_int, c_char_p, c_char_pp, c_int)),
        ('search_positions'            , CFUNCTYPE(c_int, c_char_p)),
        ('send_to_all'                 , CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('send_to_auction'             , CFUNCTYPE(c_int, c_char_p)),
        ('send_to_char'                , CFUNCTYPE(c_int, CharacterData.P, c_char_p, c_char_p)),
        ('send_to_outdoor'             , CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('set_fighting'                , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('setvalue_bufpool'            , CFUNCTYPE(c_int, DescriptorData.TxtBlock.P)),
        ('setvalue_character_list'     , CFUNCTYPE(c_int, CharacterData.P)),
        ('setvalue_loadIntoObjectIndex', CFUNCTYPE(c_int, obj_rnum)),
        ('setvalue_loadIntoZoneIndex'  , CFUNCTYPE(c_int, zone_rnum)),
        ('setvalue_mini_mud'           , CFUNCTYPE(c_int, c_int)),
        ('setvalue_obj_proto'          , CFUNCTYPE(c_int, ItemData.P)),
        ('setvalue_obj_index'          , CFUNCTYPE(c_int, IndexData.P)),
        ('setvalue_top_of_objt'        , CFUNCTYPE(c_int, obj_rnum)),
        ('setvalue_player_fl'          , CFUNCTYPE(c_int, FILE_P)),
        ('setvalue_r_frozen_start_room', CFUNCTYPE(c_int, room_rnum)),
        ('setvalue_r_immort_start_room', CFUNCTYPE(c_int, room_rnum)),
        ('setvalue_r_newbie_start_room', CFUNCTYPE(c_int, room_rnum)),
        ('setvalue_r_mortal_start_room', CFUNCTYPE(c_int, room_rnum)),
        ('setvalue_shutdown_by'        , CFUNCTYPE(c_void, c_long)),
        ('setvalue_shutdown_mode'      , CFUNCTYPE(c_void, c_int)),
        ('setvalue_shutdown_modes'     , CFUNCTYPE(c_int, c_int)),
        ('setvalue_shutdown_timer'     , CFUNCTYPE(c_void, c_int)),
        ('setvalue_olc_index'          , CFUNCTYPE(c_int, OlcIndexData.P)),
        ('setvalue_top_of_world'       , CFUNCTYPE(c_int, room_rnum)),
        ('setvalue_top_of_zone_table'  , CFUNCTYPE(c_int, zone_rnum)),
        ('setvalue_world'              , CFUNCTYPE(c_int, RoomData.P)),
        ('setvalue_zone_table'         , CFUNCTYPE(c_int, ZoneData.P)),
        ('sever_limb'                  , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('show_obj_modifiers'          , CFUNCTYPE(c_int, ItemData.P, CharacterData.P)),
        ('show_obj_to_char'            , CFUNCTYPE(c_int, ItemData.P, CharacterData.P, c_int)),
        ('shutdownMother'              , CFUNCTYPE(c_int)),
        ('shutdown_message'            , CFUNCTYPE(c_int)),
        ('spec_proc_func'              , AUTOFUNCTYPE(SPECIAL_TYPE, c_char_p, c_int)),
        ('special'                     , CFUNCTYPE(c_int, CharacterData.P, c_int, c_char_p)),
        ('spell_name'                  , CFUNCTYPE(c_char_p, c_int)),
        ('sprintbit'                   , CFUNCTYPE(c_size_t, c_ulong, c_char_pp, c_char_p, c_ulong)),
        ('sprinttype'                  , CFUNCTYPE(c_size_t, c_int, c_char_pp, c_char_p, c_ulong)),
        ('stop_fighting'               , CFUNCTYPE(c_int, CharacterData.P)),
        ('stop_hunting'                , CFUNCTYPE(c_int, CharacterData.P, CharacterData.P)),
        ('store_to_char'               , CFUNCTYPE(c_void, CharacterData.FileU.P, CharacterData.P, c_int)),
        ('string_write'                , CFUNCTYPE(c_int, DescriptorData.P, c_char_pp, c_ulong, c_long, c_void_p)),
        ('timediff'                    , CFUNCTYPE(c_int, c_timeval.P, c_timeval.P, c_timeval.P)),
        ('toggle_underworld_bridge'    , CFUNCTYPE(c_int)),
        ('unequip_char'                , AUTOFUNCTYPE(ItemData.P, CharacterData.P, c_int, c_int, c_int)),
        ('unload_dns_cache'            , CFUNCTYPE(c_int)),
        ('unlockMUD'                   , CFUNCTYPE(c_int, c_char_p)),
        ('update_index'                , CFUNCTYPE(c_int, c_char_p, c_int)),
        ('update_pos'                  , CFUNCTYPE(c_void, CharacterData.P)),
        ('update_shutdown'             , CFUNCTYPE(c_int)),
        ('valid_email'                 , CFUNCTYPE(c_int, c_char_p)),
        ('write_ban_list'              , CFUNCTYPE(c_int)),
        ('write_last_backup'           , CFUNCTYPE(c_int)),
        ('write_olc_file'              , CFUNCTYPE(c_int)),
        ('write_to_output'             , CFUNCTYPE(c_int, DescriptorData.P, c_char_p, c_char_p)),
        ('write_to_q'                  , CFUNCTYPE(c_int, c_char_p, DescriptorData.TxtQueue.P, c_int))]

MUDOperations.P = POINTER(MUDOperations)



# MUD Runtime -- Computing API
# --
# A centralized application platform core that objectifies its implementing
# components in relation to the interface layers.
#
# It also contains the instruction execution processing modules.

from .events import *
from .architecture import *

from . import registry

# Todo: Perform object registration later during boot cycle.
#@runtime.MUD.Runtime.API
class Runtime(registry.Access):
    # System classification.

    # This could be documented much better:
    # ---
    # Represents the main access object for mud runtime layer.
    # Utilizes registry access base class for addressing the
    #    named object directory.
    #
    # The actual instance is created by registry subroutines.
    # Should probably be a Singleton.

    pass

MudRuntime = Runtime


# Event Names.
EVENT_NAMES = ['bootStart'                  ,'bootComplete'               ,
               'resetStart'                 ,'resetComplete'              ,
               'greetPlayer'                ,'welcomePlayer'              ,
               'makePrompt'                 ,'loadPlayerStore'            ,
               'playerSavePoint'            ,'startRentSave'              ,
               'completeRentSave'           ,'startRentLoad'              ,
               'completeRentLoad'           ,'deleteRentFile'             ,
               'idleToVoid'                 ,'timedOut'                   ,
               'helpQueryNotFound'          ,'helpQueryUnfinished'        ,
               'helpQueryComplete'          ,'doPython'                   ,
               'createItemPrototype'        ,'createItemInstance'         ,
               'extractItemInstance'        ,'createMobilePrototype'      ,
               'createMobileInstance'       ,'extractMobileInstance'      ,
               'movement'                   ,'dealDamage'                 ,
               'slayMobile'                 ,'purgeMobile'                ,
               'deathTrap'                  ,'createRoom'                 ,
               'startZoneReset'             ,'endZoneReset'               ,
               'itemToRoom'                 ,'itemFromRoom'               ,
               'itemToContainer'            ,'itemFromContainer'          ,
               'itemToCarrier'              ,'itemFromCarrier'            ,
               'mobileToRoom'               ,'mobileFromRoom'             ,
               'lookAtRoom'                 ,'firstSpecial'               ,
               'lastSpecial'                ,'mudlog'                     ,
               'telnetCommand'              ,'writeItemPrototypeStart'    ,
               'writeItemPrototypeComplete' ,'startSpellAssign'           ,
               'startSkillAssign'           ,'newConnection'              ,
               'disconnection'              ,'enterGame'                  ,
               'quitGame'                   ,'newPlayer'                  ,
               'loadZone'                   ,'createZone'                 ,
               'castSpellSuperior'          ,'startSpellCast'             ,
               'completeSpellCast'          ,'callMagic'                  ,
               'saySpell'                   ,'lookupSkill'                ,
               'setSpellLevel'              ,'sendMail'                   ,
               'switchCutBy'                ,'unswitchForcedBy'           ,
               'usurpedBy'                  ,'duplicateKilledBy'          ,
               'reconnection'               ,'usurped'                    ,
               'passwordChange'             ,'shutdownGame'               ,
               'playerInput'                ,'playerActivation'           ,
               'performAct'                 ,'startZoneReset'             ,
               'completeZoneReset'          ,'portalOut'                  ,
               'portalIn'                   ,'manualSpell'                ,
               'libraryNode'                ,'loggedIn']

BINDINGS = \
    """
    [MemberResolution]
     # The heartbeat object is implemented directly, not as a binding.
     #  (see mud.installBridge)
     # heartbeat                   world.heartbeat.pulse # (Bridge Module)

     shutdownGame                stuphmud.server.shutdownGame

     doPython                    stuphmud.server.player.doPython
     newConnection               stuphmud.server.player.interpret

     # login                       stuphmud.server.player.login
     loggedIn                    stuphmud.server.player.loggedIn
     enterGame                   stuphmud.server.player.enterGame

     newPlayer                   stuphmud.server.player.newPlayer

     greetPlayer                 stuphmud.server.player.greetPlayer
     welcomePlayer               stuphmud.server.player.welcomePlayer

     lookAtRoom                  stuphmud.server.player.lookAtRoom

     makePrompt                  stuphmud.server.player.shell.makePrompt

     telnetCommand               stuphmud.server.player.telnet.process_command

     performAct                  stuphos.triggers.performAct
     slayMobile                  stuphos.triggers.slayMobile
     createMobileInstance        stuphos.triggers.createMobileInstance

     departure                   stuphos.triggers.departure
     arrival                     stuphos.triggers.arrival

     portalOut                   stuphos.triggers.portalOut
     portalIn                    stuphos.triggers.portalIn

     lookupSkill                 stuphmud.server.lang.magic.lookupSkill
     manualSpell                 stuphmud.server.lang.magic.manualSpell

     # writeItemPrototypeStart     stuphmud.server.zones.olc.writeItemPrototypeStart
     # writeItemPrototypeComplete  stuphmud.server.zones.olc.writeItemPrototypeComplete

    ## playerSavePoint           stuphmud.server.player.db.playerSavePoint
    ## loadPlayerStore           stuphmud.server.player.db.loadPlayerStore

    ## deleteRentFile            stuphmud.server.deleteRentFile

    ## createMobileInstance      zones.MobileCreation
    ## createItemInstance        zones.ItemCreation

    [Logging]
    ##     quitGame                
    ##     disconnection           
    ##     idleToVoid              
    ##     timedOut                
    ##
    ##     helpQueryNotFound       
    ##     helpQueryUnfinished     
    ##     helpQueryComplete       
    ##
    ##     startRentSave           
    ##     completeRentSave        
    ##     startRentLoad
    ##     completeRentLoad
    ##
    ##    ## extractMobileInstance   
    ##    ## extractItemInstance     
    ##
    ##    ## startZoneReset          
    ##    ## endZoneReset            
    ##
    ##     createMobilePrototype   
    ##     createItemPrototype     
    ##     createZone              
    ##     createRoom              
    ##
    ##    ## movement                
    ##    ## dealDamage              
    ##
    ##     slayMobile              
    ##     purgeMobile             
    ##     deathTrap               
    ##
    ##     startSpellAssign
    ##     telnetCommand
    """



# Platform-Related Routines.
PIDFILE_PATH = 'etc/process_id' # .mud_process_id

def writePidfile(pidfile = PIDFILE_PATH):
    from os import getpid

    try: print(getpid(), file=open(PIDFILE_PATH, 'w'))
    except:
        from traceback import print_exc
        print_exc()

# In-Game.
def sigusr2():
    # Emergency unrestrict.
    from os import getpid, kill
    from signal import SIGUSR2

    kill(getpid(), SIGUSR2)
